from __future__ import annotations

from collections.abc import Callable, Iterator, Sequence
from typing import TYPE_CHECKING, Any, Literal, TypeVar, overload

from safeds._utils import _structural_hash
from safeds._validation._check_columns_are_numeric import _check_column_is_numeric
from safeds.data.tabular.plotting import ColumnPlotter
from safeds.data.tabular.typing._polars_data_type import _PolarsDataType
from safeds.exceptions import (
    ColumnLengthMismatchError,
    IndexOutOfBoundsError,
    MissingValuesColumnError,
)

from ._lazy_cell import _LazyCell

if TYPE_CHECKING:
    from polars import Series

    from safeds.data.tabular.typing import DataType

    from ._cell import Cell
    from ._table import Table


T_co = TypeVar("T_co", covariant=True)
R_co = TypeVar("R_co", covariant=True)


class Column(Sequence[T_co]):
    """
    A named, one-dimensional collection of homogeneous values.

    Parameters
    ----------
    name:
        The name of the column.
    data:
        The data of the column. If None, an empty column is created.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Column
    >>> Column("test", [1, 2, 3])
    +------+
    | test |
    |  --- |
    |  i64 |
    +======+
    |    1 |
    |    2 |
    |    3 |
    +------+
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Import
    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def _from_polars_series(data: Series) -> Column:
        result = object.__new__(Column)
        result._series = data
        return result

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, name: str, data: Sequence[T_co] | None = None) -> None:
        import polars as pl

        if data is None:
            data = []

        self._series: pl.Series = pl.Series(name, data)

    def __contains__(self, item: Any) -> bool:
        return self._series.__contains__(item)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Column):
            return NotImplemented
        if self is other:
            return True
        return self._series.equals(other._series)

    @overload
    def __getitem__(self, index: int) -> T_co: ...

    @overload
    def __getitem__(self, index: slice) -> Column[T_co]: ...

    def __getitem__(self, index: int | slice) -> T_co | Column[T_co]:
        if isinstance(index, int):
            return self.get_value(index)
        else:
            start = index.start or 0
            stop = index.stop or self.row_count
            step = index.step or 1

            if start < 0 or stop < 0 or step < 0:
                raise IndexError("Negative values for start/stop/step of slices are not supported.")
            return self._from_polars_series(self._series.__getitem__(index))

    def __hash__(self) -> int:
        return _structural_hash(
            self.name,
            self.type.__repr__(),
            self.row_count,
        )

    def __iter__(self) -> Iterator[T_co]:
        return self._series.__iter__()

    def __len__(self) -> int:
        return self.row_count

    def __repr__(self) -> str:
        return self.to_table().__repr__()

    def __sizeof__(self) -> int:
        return self._series.estimated_size()

    def __str__(self) -> str:
        return self.to_table().__str__()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def is_numeric(self) -> bool:
        """Whether the column is numeric."""
        return self._series.dtype.is_numeric()

    @property
    def is_temporal(self) -> bool:
        """Whether the column is operator."""
        return self._series.dtype.is_temporal()

    @property
    def name(self) -> str:
        """The name of the column."""
        return self._series.name

    @property
    def row_count(self) -> int:
        """The number of rows in the column."""
        return self._series.len()

    @property
    def plot(self) -> ColumnPlotter:
        """The plotter for the column."""
        return ColumnPlotter(self)

    @property
    def type(self) -> DataType:
        """The type of the column."""
        return _PolarsDataType(self._series.dtype)

    # ------------------------------------------------------------------------------------------------------------------
    # Value operations
    # ------------------------------------------------------------------------------------------------------------------

    @overload
    def get_distinct_values(
        self,
        *,
        ignore_missing_values: Literal[True] = ...,
    ) -> Sequence[T_co]: ...

    @overload
    def get_distinct_values(
        self,
        *,
        ignore_missing_values: bool,
    ) -> Sequence[T_co | None]: ...

    def get_distinct_values(
        self,
        *,
        ignore_missing_values: bool = True,
    ) -> Sequence[T_co | None]:
        """
        Return the distinct values in the column.

        Parameters
        ----------
        ignore_missing_values:
            Whether to ignore missing values.

        Returns
        -------
        distinct_values:
            The distinct values in the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3, 2])
        >>> column.get_distinct_values()
        [1, 2, 3]
        """
        import polars as pl

        if self.row_count == 0:
            return []  # polars raises otherwise
        elif self._series.dtype == pl.Null:
            # polars raises otherwise
            if ignore_missing_values:
                return []
            else:
                return [None]

        if ignore_missing_values:
            series = self._series.drop_nulls()
        else:
            series = self._series

        return series.unique(maintain_order=True).to_list()

    def get_value(self, index: int) -> T_co:
        """
        Return the column value at specified index. Equivalent to the `[]` operator (indexed access).

        Nonnegative indices are counted from the beginning (starting at 0), negative indices from the end (starting at
        -1).

        Parameters
        ----------
        index:
            Index of requested value.

        Returns
        -------
        value:
            Value at index.

        Raises
        ------
        IndexError
            If the index is out of bounds.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.get_value(1)
        2

        >>> column[1]
        2
        """
        if index < -self.row_count or index >= self.row_count:
            raise IndexOutOfBoundsError(index)

        return self._series.__getitem__(index)

    # ------------------------------------------------------------------------------------------------------------------
    # Reductions
    # ------------------------------------------------------------------------------------------------------------------

    @overload
    def all(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: Literal[True] = ...,
    ) -> bool: ...

    @overload
    def all(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: bool,
    ) -> bool | None: ...

    def all(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: bool = True,
    ) -> bool | None:
        """
        Return whether all values in the column satisfy the predicate.

        The predicate can return one of three values:

        * True, if the value satisfies the predicate.
        * False, if the value does not satisfy the predicate.
        * None, if the truthiness of the predicate is unknown, e.g. due to missing values.

        By default, cases where the truthiness of the predicate is unknown are ignored and this method returns

        * True, if the predicate always returns True or None.
        * False, if the predicate returns False at least once.

        You can instead enable Kleene logic by setting `ignore_unknown=False`. In this case, this method returns

        * True, if the predicate always returns True.
        * False, if the predicate returns False at least once.
        * None, if the predicate never returns False, but at least once None.

        Parameters
        ----------
        predicate:
            The predicate to apply to each value.
        ignore_unknown:
            Whether to ignore cases where the truthiness of the predicate is unknown.

        Returns
        -------
        all_satisfy_predicate:
            Whether all values in the column satisfy the predicate.

        Raises
        ------
        TypeError
            If the predicate does not return a boolean cell.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.all(lambda cell: cell > 0)
        True

        >>> column.all(lambda cell: cell < 3)
        False
        """
        import polars as pl

        # Expressions only work on data frames/lazy frames, so we wrap the polars series first
        expression = predicate(_LazyCell(pl.col(self.name)))._polars_expression.all(ignore_nulls=ignore_unknown)
        return self._series.to_frame().select(expression).item()

    @overload
    def any(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: Literal[True] = ...,
    ) -> bool: ...

    @overload
    def any(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: bool,
    ) -> bool | None: ...

    def any(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: bool = True,
    ) -> bool | None:
        """
        Return whether any value in the column satisfies the predicate.

        The predicate can return one of three values:

        * True, if the value satisfies the predicate.
        * False, if the value does not satisfy the predicate.
        * None, if the truthiness of the predicate is unknown, e.g. due to missing values.

        By default, cases where the truthiness of the predicate is unknown are ignored and this method returns

        * True, if the predicate returns True at least once.
        * False, if the predicate always returns False or None.

        You can instead enable Kleene logic by setting `ignore_unknown=False`. In this case, this method returns

        * True, if the predicate returns True at least once.
        * False, if the predicate always returns False.
        * None, if the predicate never returns True, but at least once None.

        Parameters
        ----------
        predicate:
            The predicate to apply to each value.
        ignore_unknown:
            Whether to ignore cases where the truthiness of the predicate is unknown.

        Returns
        -------
        any_satisfy_predicate:
            Whether any value in the column satisfies the predicate.

        Raises
        ------
        TypeError
            If the predicate does not return a boolean cell.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.any(lambda cell: cell > 2)
        True

        >>> column.any(lambda cell: cell < 0)
        False
        """
        import polars as pl

        # Expressions only work on data frames/lazy frames, so we wrap the polars series first
        expression = predicate(_LazyCell(pl.col(self.name)))._polars_expression.any(ignore_nulls=ignore_unknown)
        return self._series.to_frame().select(expression).item()

    @overload
    def count_if(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: Literal[True] = ...,
    ) -> int: ...

    @overload
    def count_if(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: bool,
    ) -> int | None: ...

    def count_if(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: bool = True,
    ) -> int | None:
        """
        Return how many values in the column satisfy the predicate.

        The predicate can return one of three results:

        * True, if the value satisfies the predicate.
        * False, if the value does not satisfy the predicate.
        * None, if the truthiness of the predicate is unknown, e.g. due to missing values.

        By default, cases where the truthiness of the predicate is unknown are ignored and this method returns how
        often the predicate returns True.

        You can instead enable Kleene logic by setting `ignore_unknown=False`. In this case, this method returns None if
        the predicate returns None at least once. Otherwise, it still returns how often the predicate returns True.

        Parameters
        ----------
        predicate:
            The predicate to apply to each value.
        ignore_unknown:
            Whether to ignore cases where the truthiness of the predicate is unknown.

        Returns
        -------
        count:
            The number of values in the column that satisfy the predicate.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.count_if(lambda cell: cell > 1)
        2

        >>> column.count_if(lambda cell: cell < 0)
        0
        """
        import polars as pl

        # Expressions only work on data frames/lazy frames, so we wrap the polars series first
        expression = predicate(_LazyCell(pl.col(self.name)))._polars_expression
        series = self._series.to_frame().select(expression.alias(self.name)).get_column(self.name)

        if ignore_unknown or series.null_count() == 0:
            return series.sum()
        else:
            return None

    @overload
    def none(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: Literal[True] = ...,
    ) -> bool: ...

    @overload
    def none(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: bool,
    ) -> bool | None: ...

    def none(
        self,
        predicate: Callable[[Cell[T_co]], Cell[bool | None]],
        *,
        ignore_unknown: bool = True,
    ) -> bool | None:
        """
        Return whether no value in the column satisfies the predicate.

        The predicate can return one of three values:

        * True, if the value satisfies the predicate.
        * False, if the value does not satisfy the predicate.
        * None, if the truthiness of the predicate is unknown, e.g. due to missing values.

        By default, cases where the truthiness of the predicate is unknown are ignored and this method returns

        * True, if the predicate always returns False or None.
        * False, if the predicate returns True at least once.

        You can instead enable Kleene logic by setting `ignore_unknown=False`. In this case, this method returns

        * True, if the predicate always returns False.
        * False, if the predicate returns True at least once.
        * None, if the predicate never returns True, but at least once None.

        Parameters
        ----------
        predicate:
            The predicate to apply to each value.
        ignore_unknown:
            Whether to ignore cases where the truthiness of the predicate is unknown.

        Returns
        -------
        none_satisfy_predicate:
            Whether no value in the column satisfies the predicate.

        Raises
        ------
        TypeError
            If the predicate does not return a boolean cell.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.none(lambda cell: cell < 0)
        True

        >>> column.none(lambda cell: cell > 2)
        False
        """
        any_ = self.any(predicate, ignore_unknown=ignore_unknown)
        if any_ is None:
            return None

        return not any_

    # ------------------------------------------------------------------------------------------------------------------
    # Transformations
    # ------------------------------------------------------------------------------------------------------------------

    def rename(self, new_name: str) -> Column[T_co]:
        """
        Return a new column with a new name.

        **Note:** The original column is not modified.

        Parameters
        ----------
        new_name:
            The new name of the column.

        Returns
        -------
        renamed_column:
            A new column with the new name.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.rename("new_name")
        +----------+
        | new_name |
        |      --- |
        |      i64 |
        +==========+
        |        1 |
        |        2 |
        |        3 |
        +----------+
        """
        return self._from_polars_series(self._series.rename(new_name))

    def transform(
        self,
        transformer: Callable[[Cell[T_co]], Cell[R_co]],
    ) -> Column[R_co]:
        """
        Return a new column with values transformed by the transformer.

        **Note:** The original column is not modified.

        Parameters
        ----------
        transformer:
            The transformer to apply to each value.

        Returns
        -------
        transformed_column:
            A new column with transformed values.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.transform(lambda cell: 2 * cell)
        +------+
        | test |
        |  --- |
        |  i64 |
        +======+
        |    2 |
        |    4 |
        |    6 |
        +------+
        """
        import polars as pl

        # Expressions only work on data frames/lazy frames, so we wrap the polars series first
        expression = transformer(_LazyCell(pl.col(self.name)))._polars_expression
        series = self._series.to_frame().select(expression.alias(self.name)).get_column(self.name)

        return self._from_polars_series(series)

    # ------------------------------------------------------------------------------------------------------------------
    # Statistics
    # ------------------------------------------------------------------------------------------------------------------

    def summarize_statistics(self) -> Table:
        """
        Return a table with important statistics about the column.

        Returns
        -------
        statistics:
            The table with statistics.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("a", [1, 3])
        >>> column.summarize_statistics()
        +----------------------+---------+
        | metric               |       a |
        | ---                  |     --- |
        | str                  |     f64 |
        +================================+
        | min                  | 1.00000 |
        | max                  | 3.00000 |
        | mean                 | 2.00000 |
        | median               | 2.00000 |
        | standard deviation   | 1.41421 |
        | distinct value count | 2.00000 |
        | idness               | 1.00000 |
        | missing value ratio  | 0.00000 |
        | stability            | 0.50000 |
        +----------------------+---------+
        """
        from ._table import Table

        # TODO: turn this around (call table method, implement in table; allows parallelization)
        if self.is_numeric:
            values: list[Any] = [
                self.min(),
                self.max(),
                self.mean(),
                self.median(),
                self.standard_deviation(),
                self.distinct_value_count(),
                self.idness(),
                self.missing_value_ratio(),
                self.stability(),
            ]
        else:
            min_ = self.min()
            max_ = self.max()

            values = [
                str("-" if min_ is None else min_),
                str("-" if max_ is None else max_),
                "-",
                "-",
                "-",
                str(self.distinct_value_count()),
                str(self.idness()),
                str(self.missing_value_ratio()),
                str(self.stability()),
            ]

        return Table(
            {
                "metric": [
                    "min",
                    "max",
                    "mean",
                    "median",
                    "standard deviation",
                    "distinct value count",
                    "idness",
                    "missing value ratio",
                    "stability",
                ],
                self.name: values,
            },
        )

    def correlation_with(self, other: Column) -> float:
        """
        Calculate the Pearson correlation between this column and another column.

        The Pearson correlation is a value between -1 and 1 that indicates how much the two columns are linearly
        related:

        - A correlation of -1 indicates a perfect negative linear relationship.
        - A correlation of 0 indicates no linear relationship.
        - A correlation of 1 indicates a perfect positive linear relationship.

        Parameters
        ----------
        other:
            The other column to calculate the correlation with.

        Returns
        -------
        correlation:
            The Pearson correlation between the two columns.

        Raises
        ------
        TypeError
            If one of the columns is not numeric.
        ValueError
            If the columns have different lengths.
        ValueError
            If one of the columns has missing values.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 2, 3])
        >>> column2 = Column("test", [2, 4, 6])
        >>> column1.correlation_with(column2)
        1.0

        >>> column3 = Column("test", [3, 2, 1])
        >>> column1.correlation_with(column3)
        -1.0
        """
        import polars as pl

        _check_column_is_numeric(self, operation="calculate the correlation")
        _check_column_is_numeric(other, operation="calculate the correlation")

        if self.row_count != other.row_count:
            raise ColumnLengthMismatchError("")  # TODO: Add column names to error message
        if self.missing_value_count() > 0 or other.missing_value_count() > 0:
            raise MissingValuesColumnError("")  # TODO: Add column names to error message

        return pl.DataFrame({"a": self._series, "b": other._series}).corr().item(row=1, column="a")

    def distinct_value_count(
        self,
        *,
        ignore_missing_values: bool = True,
    ) -> int:
        """
        Return the number of distinct values in the column.

        Parameters
        ----------
        ignore_missing_values:
            Whether to ignore missing values when counting distinct values.

        Returns
        -------
        distinct_value_count:
            The number of distinct values in the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3, 2])
        >>> column.distinct_value_count()
        3
        """
        if ignore_missing_values:
            return self._series.drop_nulls().n_unique()
        else:
            return self._series.n_unique()

    def idness(self) -> float:
        """
        Calculate the idness of this column.

        We define the idness as the number of distinct values (including missing values) divided by the number of rows.
        If the column is empty, the idness is 1.0.

        A high idness indicates that the column most values in the column are unique. In this case, you must be careful
        when using the column for analysis, as a model may learn a mapping from this column to the target.

        Returns
        -------
        idness:
            The idness of the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 2, 3])
        >>> column1.idness()
        1.0

        >>> column2 = Column("test", [1, 2, 3, 2])
        >>> column2.idness()
        0.75
        """
        if self.row_count == 0:
            return 1.0  # All values are unique (since there are none)

        return self.distinct_value_count(ignore_missing_values=False) / self.row_count

    def max(self) -> T_co | None:
        """
        Return the maximum value in the column.

        Returns
        -------
        max:
            The maximum value in the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.max()
        3
        """
        from polars.exceptions import InvalidOperationError

        try:
            # Fails for Null columns
            return self._series.max()
        except InvalidOperationError:
            return None  # Return None to indicate that we don't know the maximum (consistent with mean and median)

    def mean(self) -> T_co:
        """
        Return the mean of the values in the column.

        The mean is the sum of the values divided by the number of values.

        Returns
        -------
        mean:
            The mean of the values in the column.

        Raises
        ------
        TypeError
            If the column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.mean()
        2.0
        """
        _check_column_is_numeric(self, operation="calculate the mean")

        return self._series.mean()

    def median(self) -> T_co:
        """
        Return the median of the values in the column.

        The median is the value in the middle of the sorted list of values. If the number of values is even, the median
        is the mean of the two middle values.

        Returns
        -------
        median:
            The median of the values in the column.

        Raises
        ------
        TypeError
            If the column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.median()
        2.0
        """
        _check_column_is_numeric(self, operation="calculate the median")

        return self._series.median()

    def min(self) -> T_co | None:
        """
        Return the minimum value in the column.

        Returns
        -------
        min:
            The minimum value in the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.min()
        1
        """
        from polars.exceptions import InvalidOperationError

        try:
            # Fails for Null columns
            return self._series.min()
        except InvalidOperationError:
            return None  # Return None to indicate that we don't know the maximum (consistent with mean and median)

    def missing_value_count(self) -> int:
        """
        Return the number of missing values in the column.

        Returns
        -------
        missing_value_count:
            The number of missing values in the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, None, 3])
        >>> column.missing_value_count()
        1
        """
        return self._series.null_count()

    def missing_value_ratio(self) -> float:
        """
        Return the missing value ratio.

        We define the missing value ratio as the number of missing values in the column divided by the number of rows.
        If the column is empty, the missing value ratio is 1.0.

        A high missing value ratio indicates that the column is dominated by missing values. In this case, the column
        may not be useful for analysis.

        Returns
        -------
        missing_value_ratio:
            The ratio of missing values in the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, None, 3, None])
        >>> column.missing_value_ratio()
        0.5
        """
        if self.row_count == 0:
            return 1.0  # All values are missing (since there are none)

        return self._series.null_count() / self.row_count

    @overload
    def mode(
        self,
        *,
        ignore_missing_values: Literal[True] = ...,
    ) -> Sequence[T_co]: ...

    @overload
    def mode(
        self,
        *,
        ignore_missing_values: bool,
    ) -> Sequence[T_co | None]: ...

    def mode(
        self,
        *,
        ignore_missing_values: bool = True,
    ) -> Sequence[T_co | None]:
        """
        Return the mode of the values in the column.

        The mode is the value that appears most frequently in the column. If multiple values occur equally often, all
        of them are returned. The values are sorted in ascending order.

        Parameters
        ----------
        ignore_missing_values:
            Whether to ignore missing values.

        Returns
        -------
        mode:
            The mode of the values in the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [3, 1, 2, 1, 3])
        >>> column.mode()
        [1, 3]
        """
        import polars as pl

        if self.row_count == 0:
            return []  # polars raises otherwise
        elif self._series.dtype == pl.Null:
            # polars raises otherwise
            if ignore_missing_values:
                return []
            else:
                return [None]

        if ignore_missing_values:
            series = self._series.drop_nulls()
        else:
            series = self._series

        return series.mode().sort().to_list()

    def stability(self) -> float:
        """
        Return the stability of the column.

        We define the stability as the number of occurrences of the most common non-missing value divided by the total
        number of non-missing values. If the column is empty or all values are missing, the stability is 1.0.

        A high stability indicates that the column is dominated by a single value. In this case, the column may not be
        useful for analysis.

        Returns
        -------
        stability:
            The stability of the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 1, 2, 3, None])
        >>> column.stability()
        0.5
        """
        non_missing = self._series.drop_nulls()
        if non_missing.len() == 0:
            return 1.0  # All non-null values are the same (since there is are none)

        # `unique_counts` crashes in polars for boolean columns
        mode_count = non_missing.value_counts().get_column("count").max()

        return mode_count / non_missing.len()

    def standard_deviation(self) -> float:
        """
        Return the standard deviation of the values in the column.

        The standard deviation is the square root of the variance.

        Returns
        -------
        standard_deviation:
            The standard deviation of the values in the column. If no standard deviation can be calculated due to the
            type of the column, None is returned.

        Raises
        ------
        TypeError
            If the column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.standard_deviation()
        1.0
        """
        _check_column_is_numeric(self, operation="calculate the standard deviation")

        return self._series.std()

    def variance(self) -> float:
        """
        Return the variance of the values in the column.

        The variance is the average of the squared differences from the mean.

        Raises
        ------
        TypeError
            If the column is not numeric.

        Returns
        -------
        variance:
            The variance of the values in the column. If no variance can be calculated due to the type of the column,
            None is returned.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.variance()
        1.0
        """
        _check_column_is_numeric(self, operation="calculate the variance")

        return self._series.var()

    # ------------------------------------------------------------------------------------------------------------------
    # Export
    # ------------------------------------------------------------------------------------------------------------------

    def to_list(self) -> list[T_co]:
        """
        Return the values of the column in a list.

        Returns
        -------
        values:
            The values of the column in a list.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.to_list()
        [1, 2, 3]
        """
        return self._series.to_list()

    def to_table(self) -> Table:
        """
        Create a table that contains only this column.

        Returns
        -------
        table:
            The table with this column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.to_table()
        +------+
        | test |
        |  --- |
        |  i64 |
        +======+
        |    1 |
        |    2 |
        |    3 |
        +------+
        """
        from ._table import Table

        return Table._from_polars_data_frame(self._series.to_frame())

    # ------------------------------------------------------------------------------------------------------------------
    # IPython integration
    # ------------------------------------------------------------------------------------------------------------------

    def _repr_html_(self) -> str:
        """
        Return a compact HTML representation of the column for IPython.

        Returns
        -------
        html:
            The generated HTML.
        """
        return self._series._repr_html_()



import copy
import multiprocessing as mp
from concurrent.futures import ALL_COMPLETED, wait
from typing import Self

from safeds._validation._check_columns_are_numeric import _check_columns_are_numeric
from safeds.data.labeled.containers import TabularDataset
from safeds.exceptions import (
    DatasetMissesDataError,
    FeatureDataMismatchError,
    ModelNotFittedError,
    TargetDataMismatchError,
)
from safeds.ml.classical.regression import (
    AdaBoostRegressor,
    DecisionTreeRegressor,
    ElasticNetRegressor,
    GradientBoostingRegressor,
    LassoRegressor,
    LinearRegressor,
    RandomForestRegressor,
    Regressor,
    RidgeRegressor,
    SupportVectorRegressor,
)


def _fit_single_model(model: Regressor, train_data: TabularDataset) -> Regressor:
    return model.fit(train_data)  # pragma: no cover


def _predict_single_model(model: Regressor, test_data: TabularDataset) -> TabularDataset:
    return model.predict(test_data)  # pragma: no cover


class BaselineRegressor:
    """
    Baseline Regressor.

    Get a baseline by fitting data on multiple different models and comparing the best metrics.

    Parameters
    ----------
    extended_search:
        If set to true, an extended set of models will be used to fit the classifier.
        This might result in significantly higher runtime.
    """

    def __init__(self, extended_search: bool = False):
        self._is_fitted = False
        self._list_of_model_types = [
            AdaBoostRegressor(),
            DecisionTreeRegressor(),
            LinearRegressor(),
            RandomForestRegressor(),
            RidgeRegressor(),
            SupportVectorRegressor(),
        ]

        if extended_search:
            self._list_of_model_types.extend(
                [ElasticNetRegressor(), LassoRegressor(), GradientBoostingRegressor()],
            )  # pragma: no cover

        self._fitted_models: list[Regressor] = []
        self._feature_names: list[str] | None = None
        self._target_name: str = "none"

    def fit(self, train_data: TabularDataset) -> Self:
        """
        Train the Regressor with given training data.

        The original model is not modified.

        Parameters
        ----------
        train_data:
            The data the network should be trained on.

        Returns
        -------
        trained_classifier:
            The trained Regressor

        Raises
        ------
        DatasetMissesDataError
            If the given train_data contains no data.
        ColumnTypeError
            If one or more columns contain non-numeric values.
        """
        from concurrent.futures import ProcessPoolExecutor

        # Validate Data
        train_data_as_table = train_data.to_table()
        if train_data_as_table.row_count == 0:
            raise DatasetMissesDataError
        _check_columns_are_numeric(train_data_as_table, train_data.features.add_columns(train_data.target).column_names)

        copied_model = copy.deepcopy(self)

        with ProcessPoolExecutor(
            max_workers=len(self._list_of_model_types),
            mp_context=mp.get_context("spawn"),
        ) as executor:
            futures = []
            for model in self._list_of_model_types:
                futures.append(executor.submit(_fit_single_model, model, train_data))
            [done, _] = wait(futures, return_when=ALL_COMPLETED)
            for future in done:
                copied_model._fitted_models.append(future.result())
        executor.shutdown()

        copied_model._is_fitted = True
        copied_model._feature_names = train_data.features.column_names
        copied_model._target_name = train_data.target.name
        return copied_model

    def predict(self, test_data: TabularDataset) -> dict[str, float]:
        """
        Make a prediction for the given test data and calculate the best metrics.

        The original Model is not modified.

        Parameters
        ----------
        test_data:
            The data the Regressor should predict.

        Returns
        -------
        best_metrics:
            A dictionary with the best metrics that were achieved.

        Raises
        ------
        ModelNotFittedError
            If the model has not been fitted yet
        FeatureDataMismatchError
            If the features of the test data do not match with the features of the trained Regressor.
        DatasetMissesDataError
            If the given test_data contains no data.
        TargetDataMismatchError
            If the target column of the test data does not match the target column of the training data.
        ColumnTypeError
            If one or more columns contain non-numeric values.
        """
        from concurrent.futures import ProcessPoolExecutor

        from safeds.ml.metrics import RegressionMetrics

        if not self._is_fitted:
            raise ModelNotFittedError

        # Validate data
        if not self._feature_names == test_data.features.column_names:
            raise FeatureDataMismatchError
        if not self._target_name == test_data.target.name:
            raise TargetDataMismatchError(
                actual_target_name=test_data.target.name,
                missing_target_name=self._target_name,
            )
        test_data_as_table = test_data.to_table()
        if test_data_as_table.row_count == 0:
            raise DatasetMissesDataError
        _check_columns_are_numeric(test_data_as_table, test_data.features.add_columns(test_data.target).column_names)

        # Start Processes
        with ProcessPoolExecutor(
            max_workers=len(self._list_of_model_types),
            mp_context=mp.get_context("spawn"),
        ) as executor:
            results = []
            futures = []
            for model in self._fitted_models:
                futures.append(executor.submit(_predict_single_model, model, test_data))
            [done, _] = wait(futures, return_when=ALL_COMPLETED)
            for future in done:
                results.append(future.result())
        executor.shutdown()

        # Calculate Metrics
        max_metrics = {
            "coefficient_of_determination": float("-inf"),
            "mean_absolute_error": float("inf"),
            "mean_squared_error": float("inf"),
            "median_absolute_deviation": float("inf"),
        }
        for result in results:
            coefficient_of_determination = RegressionMetrics.coefficient_of_determination(result, test_data)
            mean_absolute_error = RegressionMetrics.mean_absolute_error(result, test_data)
            mean_squared_error = RegressionMetrics.mean_squared_error(result, test_data)
            median_absolute_deviation = RegressionMetrics.median_absolute_deviation(result, test_data)

            if max_metrics.get("coefficient_of_determination", float("-inf")) < coefficient_of_determination:
                max_metrics.update({"coefficient_of_determination": coefficient_of_determination})

            if max_metrics.get("mean_absolute_error", float("inf")) > mean_absolute_error:
                max_metrics.update({"mean_absolute_error": mean_absolute_error})

            if max_metrics.get("mean_squared_error", float("inf")) > mean_squared_error:
                max_metrics.update({"mean_squared_error": mean_squared_error})

            if max_metrics.get("median_absolute_deviation", float("inf")) > median_absolute_deviation:
                max_metrics.update({"median_absolute_deviation": median_absolute_deviation})

        return max_metrics

    @property
    def is_fitted(self) -> bool:
        """Whether the model is fitted."""
        return self._is_fitted



# The class must not be nested inside a function, since pickle cannot serialize local classes. Because of this, the
# slow import of torch must be on the global level. To still evaluate the torch import lazily, the class is moved to a
# separate file.

from __future__ import annotations

from typing import TYPE_CHECKING

from torch import Tensor, nn  # slow import

from safeds._config import _init_default_device
from safeds.ml.nn.converters._input_converter_image import _InputConversionImage
from safeds.ml.nn.layers import FlattenLayer, Layer
from safeds.ml.nn.layers._pooling2d_layer import _Pooling2DLayer

if TYPE_CHECKING:
    from safeds.ml.nn.converters import InputConversion
    from safeds.ml.nn.typing import ModelImageSize


# Use torch.compile once the following issues are resolved:
# - https://github.com/pytorch/pytorch/issues/120233 (Python 3.12 support)
# - https://github.com/triton-lang/triton/issues/1640 (Windows support)
class _InternalModel(nn.Module):
    def __init__(self, input_conversion: InputConversion, layers: list[Layer], is_for_classification: bool) -> None:
        super().__init__()

        _init_default_device()

        self._layer_list = layers
        internal_layers = []
        previous_output_size = input_conversion._data_size

        for layer in layers:
            if previous_output_size is not None:
                layer._set_input_size(previous_output_size)
            elif isinstance(input_conversion, _InputConversionImage):
                layer._set_input_size(input_conversion._data_size)
            if isinstance(layer, FlattenLayer | _Pooling2DLayer):
                internal_layers.append(layer._get_internal_layer())
            else:
                internal_layers.append(layer._get_internal_layer(activation_function="relu"))
            previous_output_size = layer.output_size

        if is_for_classification:
            internal_layers.pop()
            if isinstance(layers[-1].output_size, int) and layers[-1].output_size > 2:
                internal_layers.append(layers[-1]._get_internal_layer(activation_function="none"))
            else:
                internal_layers.append(layers[-1]._get_internal_layer(activation_function="sigmoid"))
        self._pytorch_layers = nn.Sequential(*internal_layers)

    @property
    def input_size(self) -> int | ModelImageSize:
        return self._layer_list[0].input_size

    def forward(self, x: Tensor) -> Tensor:
        for layer in self._pytorch_layers:
            x = layer(x)
        return x



from __future__ import annotations

import io
import sys
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

from safeds._config import _get_device, _init_default_device
from safeds._utils import _structural_hash
from safeds.data.image._utils._image_transformation_error_and_warning_checks import (
    _check_add_noise_errors,
    _check_adjust_brightness_errors_and_warnings,
    _check_adjust_color_balance_errors_and_warnings,
    _check_adjust_contrast_errors_and_warnings,
    _check_blur_errors_and_warnings,
    _check_crop_errors,
    _check_crop_warnings,
    _check_resize_errors,
    _check_sharpen_errors_and_warnings,
)
from safeds.data.image.typing import ImageSize
from safeds.exceptions import IllegalFormatError

if TYPE_CHECKING:
    from numpy import dtype, ndarray
    from torch import Tensor


class Image:
    """
    A container for image data.

    Parameters
    ----------
    image_tensor:
        The image data as tensor.
    """

    _filter_edges_kernel_cache: Tensor | None = None

    @staticmethod
    def _filter_edges_kernel() -> Tensor:
        import torch

        _init_default_device()

        if Image._filter_edges_kernel_cache is None:
            Image._filter_edges_kernel_cache = (
                torch.tensor([[-1.0, -1.0, -1.0], [-1.0, 8.0, -1.0], [-1.0, -1.0, -1.0]], dtype=torch.float16)
                .unsqueeze(dim=0)
                .unsqueeze(dim=0)
                .to(_get_device())
            )
        if Image._filter_edges_kernel_cache.device != _get_device():
            Image._filter_edges_kernel_cache = Image._filter_edges_kernel_cache.to(_get_device())  # pragma: no cover

        return Image._filter_edges_kernel_cache

    @staticmethod
    def from_file(path: str | Path) -> Image:
        """
        Create an image from a file.

        Parameters
        ----------
        path:
            The path to the image file.

        Returns
        -------
        image:
            The image.

        Raises
        ------
        FileNotFoundError
            If the file of the path cannot be found
        """
        from torchvision.io import read_image

        if isinstance(path, str):
            path = Path(path)

        _init_default_device()

        if not path.is_file():
            raise FileNotFoundError(f"No such file or directory: '{path}'")

        return Image(image_tensor=read_image(str(path)).to(_get_device()))

    @staticmethod
    def from_bytes(data: bytes) -> Image:
        """
        Create an image from bytes.

        Parameters
        ----------
        data:
            The data of the image.

        Returns
        -------
        image:
            The image.
        """
        import torch
        import torchvision

        _init_default_device()

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="The given buffer is not writable, and PyTorch does not support non-writable tensors.",
            )
            input_tensor = torch.frombuffer(data, dtype=torch.uint8)

        return Image(image_tensor=torchvision.io.decode_image(input_tensor).to(_get_device()))

    def __init__(self, image_tensor: Tensor) -> None:
        import torch

        self._image_tensor: Tensor
        if image_tensor.dtype != torch.uint8:
            self._image_tensor = torch.clamp(image_tensor, 0, 255).to(torch.uint8)
        else:
            self._image_tensor = image_tensor

    def __eq__(self, other: object) -> bool:
        """
        Compare two images.

        Parameters
        ----------
        other:
            The image to compare to.

        Returns
        -------
        equals:
            Whether the two images contain equal pixel data.
        """
        import torch

        _init_default_device()

        if not isinstance(other, Image):
            return NotImplemented
        return (self is other) or (
            self._image_tensor.size() == other._image_tensor.size()
            and torch.all(torch.eq(self._image_tensor, other._image_tensor)).item()
        )

    def __hash__(self) -> int:
        """
        Return a deterministic hash value for this image.

        Returns
        -------
        hash:
            The hash value.
        """
        return _structural_hash(self.width, self.height, self.channel)

    def __sizeof__(self) -> int:
        """
        Return the complete size of this object.

        Returns
        -------
        size:
            Size of this object in bytes.
        """
        return sys.getsizeof(self._image_tensor) + self._image_tensor.element_size() * self._image_tensor.nelement()

    def __array__(self, numpy_dtype: str | dtype | None = None) -> ndarray:
        """
        Return the image as a numpy array.

        Returns
        -------
        numpy_array:
            The image as numpy array.
        """
        from numpy import uint8

        return (
            self._image_tensor.permute(1, 2, 0)
            .detach()
            .cpu()
            .numpy()
            .astype(uint8 if numpy_dtype is None else numpy_dtype)
        )

    def _repr_jpeg_(self) -> bytes | None:
        """
        Return a JPEG image as bytes.

        If the image has an alpha channel return None.

        Returns
        -------
        jpeg:
            The image as JPEG.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2
        from torchvision.utils import save_image

        _init_default_device()

        if self.channel == 4:
            return None
        buffer = io.BytesIO()
        if self.channel == 1:
            func2.to_pil_image(self._image_tensor, mode="L").save(buffer, format="jpeg")
        else:
            save_image(self._image_tensor.to(torch.float32) / 255, buffer, format="jpeg")
        buffer.seek(0)
        return buffer.read()

    def _repr_png_(self) -> bytes:
        """
        Return a PNG image as bytes.

        Returns
        -------
        png:
            The image as PNG.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2
        from torchvision.utils import save_image

        _init_default_device()

        buffer = io.BytesIO()
        if self.channel == 1:
            func2.to_pil_image(self._image_tensor, mode="L").save(buffer, format="png")
        else:
            save_image(self._image_tensor.to(torch.float32) / 255, buffer, format="png")
        buffer.seek(0)
        return buffer.read()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def width(self) -> int:
        """
        Get the width of the image in pixels.

        Returns
        -------
        width:
            The width of the image.
        """
        return self._image_tensor.size(dim=2)

    @property
    def height(self) -> int:
        """
        Get the height of the image in pixels.

        Returns
        -------
        height:
            The height of the image.
        """
        return self._image_tensor.size(dim=1)

    @property
    def channel(self) -> int:
        """
        Get the number of channels of the image.

        Returns
        -------
        channel:
            The number of channels of the image.
        """
        return self._image_tensor.size(dim=0)

    @property
    def size(self) -> ImageSize:
        """
        Get the `ImageSize` of the image.

        Returns
        -------
        image_size:
            The size of the image.
        """
        return ImageSize(self.width, self.height, self.channel)

    # ------------------------------------------------------------------------------------------------------------------
    # Conversion
    # ------------------------------------------------------------------------------------------------------------------

    def to_jpeg_file(self, path: str | Path) -> None:
        """
        Save the image as a JPEG file.

        Parameters
        ----------
        path:
            The path to the JPEG file.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2
        from torchvision.utils import save_image

        _init_default_device()

        if self.channel == 4:
            raise IllegalFormatError("png")
        Path(path).parent.mkdir(parents=True, exist_ok=True)
        if self.channel == 1:
            func2.to_pil_image(self._image_tensor, mode="L").save(path, format="jpeg")
        else:
            save_image(self._image_tensor.to(torch.float32) / 255, path, format="jpeg")

    def to_png_file(self, path: str | Path) -> None:
        """
        Save the image as a PNG file.

        Parameters
        ----------
        path:
            The path to the PNG file.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2
        from torchvision.utils import save_image

        _init_default_device()

        Path(path).parent.mkdir(parents=True, exist_ok=True)
        if self.channel == 1:
            func2.to_pil_image(self._image_tensor, mode="L").save(path, format="png")
        else:
            save_image(self._image_tensor.to(torch.float32) / 255, path, format="png")

    # ------------------------------------------------------------------------------------------------------------------
    # Transformations
    # ------------------------------------------------------------------------------------------------------------------

    def change_channel(self, channel: int) -> Image:
        """
        Return a new `Image` that has the given number of channels.

        The original image is not modified.

        Parameters
        ----------
        channel:
            The new number of channels. 1 will result in a grayscale image.

        Returns
        -------
        result:
            The image with the given number of channels.

        Raises
        ------
        ValueError
            if the given channel is not a valid channel option
        """
        import torch

        _init_default_device()

        if self.channel == channel:
            image_tensor = self._image_tensor
        elif self.channel == 1 and channel == 3:
            image_tensor = torch.cat([self._image_tensor, self._image_tensor, self._image_tensor], dim=0)
        elif self.channel == 1 and channel == 4:
            image_tensor = torch.cat(
                [
                    self._image_tensor,
                    self._image_tensor,
                    self._image_tensor,
                    torch.full(self._image_tensor.size(), 255),
                ],
                dim=0,
            )
        elif self.channel in (3, 4) and channel == 1:
            image_tensor = self.convert_to_grayscale()._image_tensor[0:1]
        elif self.channel == 3 and channel == 4:
            image_tensor = torch.cat(
                [self._image_tensor, torch.full(self._image_tensor[0:1].size(), 255)],
                dim=0,
            )
        elif self.channel == 4 and channel == 3:
            image_tensor = self._image_tensor[0:3]
        else:
            raise ValueError(f"Channel {channel} is not a valid channel option. Use either 1, 3 or 4")
        return Image(image_tensor)

    def resize(self, new_width: int, new_height: int) -> Image:
        """
        Return a new `Image` that has been resized to a given size.

        The original image is not modified.

        Parameters
        ----------
        new_width:
            the new width of the image
        new_height:
            the new height of the image

        Returns
        -------
        result:
            The image with the given width and height.

        Raises
        ------
        OutOfBoundsError
            If new_width or new_height are below 1
        """
        from torchvision.transforms import InterpolationMode
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_resize_errors(new_width, new_height)
        return Image(
            func2.resize(self._image_tensor, size=[new_height, new_width], interpolation=InterpolationMode.NEAREST),
        )

    def convert_to_grayscale(self) -> Image:
        """
        Return a new `Image` that is converted to grayscale.

        The new image will have the same amount of channels as the original image.
        If you want to change the amount of channels used, please use the method [change_channel][safeds.data.image.containers._image.Image.change_channel].

        The original image is not modified.

        Returns
        -------
        result:
            The grayscale image.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        if self.channel == 1:
            return self
        elif self.channel == 4:
            return Image(
                torch.cat(
                    [
                        func2.rgb_to_grayscale(self._image_tensor[0:3], num_output_channels=3),
                        self._image_tensor[3:4],
                    ],
                ),
            )
        else:  # channel == 3
            return Image(
                func2.rgb_to_grayscale(self._image_tensor[0:3], num_output_channels=3),
            )

    def crop(self, x: int, y: int, width: int, height: int) -> Image:
        """
        Return a new `Image` that has been cropped to a given bounding rectangle.

        The original image is not modified.

        Parameters
        ----------
        x:
            the x coordinate of the top-left corner of the bounding rectangle
        y:
            the y coordinate of the top-left corner of the bounding rectangle
        width:
            the width of the bounding rectangle
        height:
            the height of the bounding rectangle

        Returns
        -------
        result:
            The cropped image.

        Raises
        ------
        OutOfBoundsError
            If x or y are below 0 or if width or height are below 1
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_crop_errors(x, y, width, height)
        _check_crop_warnings(x, y, self.width, self.height, plural=False)
        return Image(func2.crop(self._image_tensor, y, x, height, width))

    def flip_vertically(self) -> Image:
        """
        Return a new `Image` that is flipped vertically (horizontal axis, flips up-down and vice versa).

        The original image is not modified.

        Returns
        -------
        result:
            The flipped image.
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        return Image(func2.vertical_flip(self._image_tensor))

    def flip_horizontally(self) -> Image:
        """
        Return a new `Image` that is flipped horizontally (vertical axis, flips left-right and vice versa).

        The original image is not modified.

        Returns
        -------
        result:
            The flipped image.
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        return Image(func2.horizontal_flip(self._image_tensor))

    def adjust_brightness(self, factor: float) -> Image:
        """
        Return a new `Image` with an adjusted brightness.

        The original image is not modified.

        Parameters
        ----------
        factor:
            The brightness factor.
            1.0 will not change the brightness.
            Below 1.0 will result in a darker image.
            Above 1.0 will resolut in a brighter image.
            Has to be bigger than or equal to 0 (black).

        Returns
        -------
        result:
            The Image with adjusted brightness.

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """
        import torch

        _init_default_device()

        _check_adjust_brightness_errors_and_warnings(factor, plural=False)
        if self._image_tensor.size(dim=-3) != 4:
            if factor == 0:
                return Image(torch.zeros(self._image_tensor.size(), dtype=torch.uint8))
            else:
                temp_tensor = self._image_tensor * torch.tensor([factor * 1.0], dtype=torch.float16)
                torch.clamp(temp_tensor, 0, 255, out=temp_tensor)
                return Image(temp_tensor.to(torch.uint8))
        else:
            img_tensor = torch.empty(self._image_tensor.size(), dtype=torch.uint8)
            img_tensor[3] = self._image_tensor[3]
            if factor == 0:
                torch.zeros(
                    (3, self._image_tensor.size(dim=-2), self._image_tensor.size(dim=-1)),
                    dtype=torch.uint8,
                    out=img_tensor[:, 0:3],
                )
            else:
                temp_tensor = self._image_tensor[0:3] * torch.tensor([factor * 1.0], dtype=torch.float16)
                torch.clamp(temp_tensor, 0, 255, out=temp_tensor)
                img_tensor[0:3] = temp_tensor[:]
            return Image(img_tensor)

    def add_noise(self, standard_deviation: float) -> Image:
        """
        Return a new `Image` with noise added to the image.

        The original image is not modified.

        Parameters
        ----------
        standard_deviation:
            The standard deviation of the normal distribution. Has to be bigger than or equal to 0.

        Returns
        -------
        result:
            The image with added noise.

        Raises
        ------
        OutOfBoundsError
            If standard_deviation is smaller than 0.
        """
        import torch

        _init_default_device()

        _check_add_noise_errors(standard_deviation)
        float_tensor = torch.empty(self._image_tensor.size(), dtype=torch.float16)
        torch.normal(0, standard_deviation, self._image_tensor.size(), out=float_tensor)
        float_tensor *= 255
        float_tensor += self._image_tensor
        torch.clamp(float_tensor, 0, 255, out=float_tensor)
        return Image(float_tensor.to(torch.uint8))

    def adjust_contrast(self, factor: float) -> Image:
        """
        Return a new `Image` with adjusted contrast.

        The original image is not modified.

        Parameters
        ----------
        factor:
            If factor > 1, increase contrast of image.
            If factor = 1, no changes will be made.
            If factor < 1, make image greyer.
            Has to be bigger than or equal to 0 (gray).

        Returns
        -------
        image:
            New image with adjusted contrast.

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """
        import torch

        _init_default_device()

        _check_adjust_contrast_errors_and_warnings(factor, plural=False)

        factor *= 1.0
        adjusted_factor = (1 - factor) / factor
        gray_tensor = self.convert_to_grayscale()._image_tensor[0]
        mean = torch.mean(gray_tensor, dim=(-2, -1), dtype=torch.float16)
        del gray_tensor
        mean *= torch.tensor(adjusted_factor, dtype=torch.float16)
        tensor = mean.repeat(min(self.channel, 3), self._image_tensor.size(dim=-2), self._image_tensor.size(dim=-1))
        tensor += self._image_tensor[0 : min(self.channel, 3)]
        tensor *= factor
        torch.clamp(tensor, 0, 255, out=tensor)

        if self.channel == 4:
            return Image(torch.cat([tensor.to(torch.uint8), self._image_tensor[3:4]], dim=0))
        else:
            return Image(tensor.to(torch.uint8))

    def adjust_color_balance(self, factor: float) -> Image:
        """
        Return a new `Image` with adjusted color balance.

        The original image is not modified.

        Parameters
        ----------
        factor:
            Has to be bigger than or equal to 0.
            If 0 <= factor < 1, make image greyer.
            If factor = 1, no changes will be made.
            If factor > 1, increase color balance of image.

        Returns
        -------
        image:
            The new, adjusted image.

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """
        import torch

        _check_adjust_color_balance_errors_and_warnings(factor, self.channel, plural=False)

        factor *= 1.0
        if factor == 0:
            return self.convert_to_grayscale()
        else:
            adjusted_factor = (1 - factor) / factor
            tensor = self.convert_to_grayscale()._image_tensor * torch.tensor(adjusted_factor, dtype=torch.float16)
            tensor += self._image_tensor
            tensor *= factor
            torch.clamp(tensor, 0, 255, out=tensor)
            return Image(tensor.to(torch.uint8))

    def blur(self, radius: int) -> Image:
        """
        Return a blurred version of the image.

        The original image is not modified.

        Parameters
        ----------
        radius:
             Radius is directly proportional to the blur value. The radius is equal to the amount of pixels united in
             each direction. A radius of 1 will result in a united box of 9 pixels.

        Returns
        -------
        result:
            The blurred image.

        Raises
        ------
        OutOfBoundsError
            If radius is smaller than 0 or equal or greater than the smaller size of the image.
        """
        import torch

        _init_default_device()

        float_dtype = torch.float32 if _get_device() != torch.device("cuda") else torch.float16

        _check_blur_errors_and_warnings(radius, min(self.width, self.height), plural=False)

        kernel = torch.full(
            (self._image_tensor.size(dim=-3), 1, radius * 2 + 1, radius * 2 + 1),
            1 / (radius * 2 + 1) ** 2,
            dtype=float_dtype,
        )
        tensor = torch.nn.functional.conv2d(
            torch.nn.functional.pad(
                self._image_tensor.to(float_dtype),
                (radius, radius, radius, radius),
                mode="replicate",
            ),
            kernel,
            padding="valid",
            groups=self._image_tensor.size(dim=-3),
        ).to(torch.uint8)
        return Image(tensor)

    def sharpen(self, factor: float) -> Image:
        """
        Return a sharpened version of the image.

        The original image is not modified.

        Parameters
        ----------
        factor:
            If factor > 1, increase the sharpness of the image.
            If factor = 1, no changes will be made.
            If factor < 1, blur the image.
            Has to be bigger than or equal to 0 (blurred).

        Returns
        -------
        result:
            The image sharpened by the given factor.

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        _check_sharpen_errors_and_warnings(factor, plural=False)
        if self.channel == 4:
            return Image(
                torch.cat(
                    [
                        func2.adjust_sharpness(self._image_tensor[0:3], factor * 1.0),
                        self._image_tensor[3:4],
                    ],
                ),
            )
        else:
            return Image(func2.adjust_sharpness(self._image_tensor, factor * 1.0))

    def invert_colors(self) -> Image:
        """
        Return a new `Image` with colors inverted.

        The original image is not modified.

        Returns
        -------
        result:
            The image with inverted colors.
        """
        import torch
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        if self.channel == 4:
            return Image(
                torch.cat([func2.invert(self._image_tensor[0:3]), self._image_tensor[3:4]]),
            )
        else:
            return Image(func2.invert(self._image_tensor))

    def rotate_right(self) -> Image:
        """
        Return a new `Image` that is rotated 90 degrees clockwise.

        The original image is not modified.

        Returns
        -------
        result:
            The image rotated 90 degrees clockwise.
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        return Image(func2.rotate(self._image_tensor, -90, expand=True))

    def rotate_left(self) -> Image:
        """
        Return a new `Image` that is rotated 90 degrees counter-clockwise.

        The original image is not modified.

        Returns
        -------
        result:
            The image rotated 90 degrees counter-clockwise.
        """
        from torchvision.transforms.v2 import functional as func2

        _init_default_device()

        return Image(func2.rotate(self._image_tensor, 90, expand=True))

    def find_edges(self) -> Image:
        """
        Return a grayscale version of the image with the edges highlighted.

        The original image is not modified.

        Returns
        -------
        result:
            The image with edges found.
        """
        import torch

        _init_default_device()

        edges_tensor_float16 = torch.nn.functional.conv2d(
            self.convert_to_grayscale()._image_tensor.to(torch.float16)[0:1],
            Image._filter_edges_kernel(),
            padding="same",
        )
        torch.clamp(edges_tensor_float16, 0, 255, out=edges_tensor_float16)
        if self.channel == 1:
            return Image(edges_tensor_float16.to(torch.uint8))
        edges_tensor = edges_tensor_float16.to(torch.uint8)
        del edges_tensor_float16
        if self.channel == 3:
            return Image(edges_tensor.repeat(3, 1, 1))
        else:  # self.channel == 4
            return Image(
                torch.cat([edges_tensor.repeat(3, 1, 1), self._image_tensor[3:4]]),
            )



from __future__ import annotations

import warnings
from typing import Any

from safeds._utils import _structural_hash
from safeds._validation import _check_columns_exist
from safeds._validation._check_columns_are_numeric import _check_columns_are_numeric
from safeds.data.tabular.containers import Table
from safeds.exceptions import TransformerNotFittedError

from ._invertible_table_transformer import InvertibleTableTransformer


class LabelEncoder(InvertibleTableTransformer):
    """
    The LabelEncoder encodes one or more given columns into labels.

    Parameters
    ----------
    column_names:
        The list of columns used to fit the transformer. If `None`, all non-numeric columns are used.
    partial_order:
        The partial order of the labels. The labels are encoded in the order of the given list. Additional values are
        assigned labels in the order they are encountered during fitting.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        *,
        column_names: str | list[str] | None = None,
        partial_order: list[Any] | None = None,
    ) -> None:
        super().__init__(column_names)

        if partial_order is None:
            partial_order = []

        # Parameters
        self._partial_order = partial_order

        # Internal state
        self._mapping: dict[str, dict[Any, int]] | None = None  # Column name -> value -> label
        self._inverse_mapping: dict[str, dict[int, Any]] | None = None  # Column name -> label -> value

    def __hash__(self) -> int:
        return _structural_hash(
            super().__hash__(),
            self._partial_order,
            # Leave out the internal state for faster hashing
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def is_fitted(self) -> bool:
        """Whether the transformer is fitted."""
        return self._mapping is not None and self._inverse_mapping is not None

    @property
    def partial_order(self) -> list[Any]:
        """The partial order of the labels."""
        return list(self._partial_order)  # defensive copy

    # ------------------------------------------------------------------------------------------------------------------
    # Learning and transformation
    # ------------------------------------------------------------------------------------------------------------------

    def fit(self, table: Table) -> LabelEncoder:
        """
        Learn a transformation for a set of columns in a table.

        This transformer is not modified.

        Parameters
        ----------
        table:
            The table used to fit the transformer.

        Returns
        -------
        fitted_transformer:
            The fitted transformer.

        Raises
        ------
        ColumnNotFoundError
            If column_names contain a column name that is missing in the table.
        ValueError
            If the table contains 0 rows.
        """
        if self._column_names is None:
            column_names = [name for name in table.column_names if not table.get_column_type(name).is_numeric]
        else:
            column_names = self._column_names
            _check_columns_exist(table, column_names)
            _warn_if_columns_are_numeric(table, column_names)

        if table.row_count == 0:
            raise ValueError("The LabelEncoder cannot be fitted because the table contains 0 rows")

        # Learn the transformation
        mapping = {}
        reverse_mapping = {}

        for name in column_names:
            # Remember partial order
            mapping[name] = {value: index for index, value in enumerate(self._partial_order)}
            reverse_mapping[name] = {index: value for value, index in mapping[name].items()}

            unique_values = table.get_column(name).get_distinct_values()
            for value in unique_values:
                if value not in mapping[name]:
                    label = len(mapping[name])
                    mapping[name][value] = label
                    reverse_mapping[name][label] = value

        # Create a copy with the learned transformation
        result = LabelEncoder(column_names=column_names, partial_order=self._partial_order)
        result._mapping = mapping
        result._inverse_mapping = reverse_mapping

        return result

    def transform(self, table: Table) -> Table:
        """
        Apply the learned transformation to a table.

        The table is not modified.

        Parameters
        ----------
        table:
            The table to which the learned transformation is applied.

        Returns
        -------
        transformed_table:
            The transformed table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        ColumnNotFoundError
            If the input table does not contain all columns used to fit the transformer.
        ValueError
            If the table contains 0 rows.
        """
        import polars as pl

        # Used in favor of is_fitted, so the type checker is happy
        if self._column_names is None or self._mapping is None:
            raise TransformerNotFittedError

        _check_columns_exist(table, self._column_names)

        columns = [
            pl.col(name).replace(self._mapping[name], default=None, return_dtype=pl.UInt32)
            for name in self._column_names
        ]

        return Table._from_polars_lazy_frame(
            table._lazy_frame.with_columns(columns),
        )

    def inverse_transform(self, transformed_table: Table) -> Table:
        """
        Undo the learned transformation.

        The table is not modified.

        Parameters
        ----------
        transformed_table:
            The table to be transformed back to the original version.

        Returns
        -------
        original_table:
            The original table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        ColumnNotFoundError
            If the input table does not contain all columns used to fit the transformer.
        ColumnTypeError
            If the specified columns of the input table contain non-numerical data.
        """
        import polars as pl

        # Used in favor of is_fitted, so the type checker is happy
        if self._column_names is None or self._inverse_mapping is None:
            raise TransformerNotFittedError

        _check_columns_exist(transformed_table, self._column_names)
        _check_columns_are_numeric(
            transformed_table,
            self._column_names,
            operation="inverse-transform with a LabelEncoder",
        )

        columns = [pl.col(name).replace(self._inverse_mapping[name], default=None) for name in self._column_names]

        return Table._from_polars_lazy_frame(
            transformed_table._lazy_frame.with_columns(columns),
        )


def _warn_if_columns_are_numeric(table: Table, column_names: list[str]) -> None:
    numeric_columns = table.remove_columns_except(column_names).remove_non_numeric_columns().column_names
    if numeric_columns:
        warnings.warn(
            f"The columns {numeric_columns} contain numerical data. "
            "The LabelEncoder is designed to encode non-numerical values into numerical values",
            UserWarning,
            stacklevel=2,
        )



"""Classes for transforming tabular data."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._discretizer import Discretizer
    from ._functional_table_transformer import FunctionalTableTransformer
    from ._invertible_table_transformer import InvertibleTableTransformer
    from ._k_nearest_neighbors_imputer import KNearestNeighborsImputer
    from ._label_encoder import LabelEncoder
    from ._one_hot_encoder import OneHotEncoder
    from ._range_scaler import RangeScaler
    from ._robust_scaler import RobustScaler
    from ._sequential_table_transformer import SequentialTableTransformer
    from ._simple_imputer import SimpleImputer
    from ._standard_scaler import StandardScaler
    from ._table_transformer import TableTransformer


apipkg.initpkg(
    __name__,
    {
        "Discretizer": "._discretizer:Discretizer",
        "FunctionalTableTransformer": "._functional_table_transformer:FunctionalTableTransformer",
        "InvertibleTableTransformer": "._invertible_table_transformer:InvertibleTableTransformer",
        "LabelEncoder": "._label_encoder:LabelEncoder",
        "OneHotEncoder": "._one_hot_encoder:OneHotEncoder",
        "RangeScaler": "._range_scaler:RangeScaler",
        "SequentialTableTransformer": "._sequential_table_transformer:SequentialTableTransformer",
        "RobustScaler": "._robust_scaler:RobustScaler",
        "SimpleImputer": "._simple_imputer:SimpleImputer",
        "StandardScaler": "._standard_scaler:StandardScaler",
        "TableTransformer": "._table_transformer:TableTransformer",
        "KNearestNeighborsImputer": "._k_nearest_neighbors_imputer:KNearestNeighborsImputer",
    },
)

__all__ = [
    "Discretizer",
    "FunctionalTableTransformer",
    "InvertibleTableTransformer",
    "LabelEncoder",
    "OneHotEncoder",
    "RangeScaler",
    "SequentialTableTransformer",
    "RobustScaler",
    "SimpleImputer",
    "StandardScaler",
    "TableTransformer",
    "KNearestNeighborsImputer",
]



"""Validation of preconditions."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._check_bounds import _check_bounds, _ClosedBound, _OpenBound
    from ._check_columns_exist import _check_columns_exist
    from ._normalize_and_check_file_path import _normalize_and_check_file_path

apipkg.initpkg(
    __name__,
    {
        "_check_bounds": "._check_bounds:_check_bounds",
        "_ClosedBound": "._check_bounds:_ClosedBound",
        "_OpenBound": "._check_bounds:_OpenBound",
        "_check_columns_exist": "._check_columns_exist:_check_columns_exist",
        "_normalize_and_check_file_path": "._normalize_and_check_file_path:_normalize_and_check_file_path",
    },
)

__all__ = [
    "_check_bounds",
    "_ClosedBound",
    "_OpenBound",
    "_check_columns_exist",
    "_normalize_and_check_file_path",
]



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Self

from safeds._utils import _structural_hash
from safeds.data.image._utils._image_transformation_error_and_warning_checks import (
    _check_add_noise_errors,
    _check_adjust_brightness_errors_and_warnings,
    _check_adjust_color_balance_errors_and_warnings,
    _check_adjust_contrast_errors_and_warnings,
    _check_blur_errors_and_warnings,
    _check_crop_errors,
    _check_remove_images_with_size_errors,
    _check_resize_errors,
    _check_sharpen_errors_and_warnings,
)
from safeds.data.image.containers._image_list import ImageList
from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
from safeds.exceptions import IndexOutOfBoundsError

if TYPE_CHECKING:
    from pathlib import Path

    from torch import Tensor

    from safeds.data.image.containers import Image
    from safeds.data.image.typing import ImageSize


class _EmptyImageList(ImageList):
    """
    An ImageList is a list of different images. It can hold different sizes of Images. The channel of all images is the same.

    This is the class for an empty ImageList.

    To create an `ImageList` call one of the following static methods:

    | Method                                                                        | Description                                              |
    | ----------------------------------------------------------------------------- | -------------------------------------------------------- |
    | [from_images][safeds.data.image.containers._image_list.ImageList.from_images] | Create an ImageList from a list of Images.               |
    | [from_files][safeds.data.image.containers._image_list.ImageList.from_files]   | Create an ImageList from a directory or a list of files. |
    """

    _instance = None

    @staticmethod
    def _warn_empty_image_list() -> None:
        """Warn if a transform method is used on an empty ImageList."""
        warnings.warn(
            "You are using an empty ImageList. This method changes nothing if used on an empty ImageList.",
            UserWarning,
            stacklevel=2,
        )

    def __new__(cls) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @staticmethod
    def _create_image_list(images: list[Tensor], indices: list[int]) -> ImageList:
        raise NotImplementedError

    def _clone(self) -> ImageList:
        return _EmptyImageList()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ImageList):
            return NotImplemented
        return isinstance(other, _EmptyImageList)

    def __hash__(self) -> int:
        return _structural_hash("_EmptyImageList")

    def __sizeof__(self) -> int:
        return 0

    @property
    def image_count(self) -> int:
        return 0

    @property
    def widths(self) -> list[int]:
        return []

    @property
    def heights(self) -> list[int]:
        return []

    @property
    def channel(self) -> int:
        return NotImplemented

    @property
    def sizes(self) -> list[ImageSize]:
        return []

    @property
    def size_count(self) -> int:
        return 0

    def get_image(self, index: int) -> Image:
        raise IndexOutOfBoundsError(index)

    def index(self, _image: Image) -> list[int]:
        return []

    def has_image(self, _image: Image) -> bool:
        return False

    def to_jpeg_files(self, _path: str | Path | list[str | Path]) -> None:
        warnings.warn("You are using an empty ImageList. No files will be saved.", UserWarning, stacklevel=2)

    def to_png_files(self, _path: str | Path | list[str | Path]) -> None:
        warnings.warn("You are using an empty ImageList. No files will be saved.", UserWarning, stacklevel=2)

    def to_images(self, _indices: list[int] | None = None) -> list[Image]:
        return []

    def change_channel(self, channel: int) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        if channel not in (1, 3, 4):
            raise ValueError(f"Channel {channel} is not a valid channel option. Use either 1, 3 or 4")
        return _EmptyImageList()

    def _add_image_tensor(self, image_tensor: Tensor, _index: int) -> ImageList:
        return _SingleSizeImageList._create_image_list([image_tensor], [0])

    def add_images(self, images: list[Image] | ImageList) -> ImageList:
        return ImageList.from_images(images) if isinstance(images, list) else images

    def remove_image_by_index(self, index: int | list[int]) -> ImageList:
        raise IndexOutOfBoundsError(index)

    def _remove_image_by_index_ignore_invalid(self, _index: int | list[int]) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        return _EmptyImageList()

    def remove_images_with_size(self, width: int, height: int) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        _check_remove_images_with_size_errors(width, height)
        return _EmptyImageList()

    def remove_duplicate_images(self) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        return _EmptyImageList()

    def shuffle_images(self) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        return _EmptyImageList()

    def resize(self, new_width: int, new_height: int) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        _check_resize_errors(new_width, new_height)
        return _EmptyImageList()

    def convert_to_grayscale(self) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        return _EmptyImageList()

    def crop(self, x: int, y: int, width: int, height: int) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        _check_crop_errors(x, y, width, height)
        return _EmptyImageList()

    def flip_vertically(self) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        return _EmptyImageList()

    def flip_horizontally(self) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        return _EmptyImageList()

    def adjust_brightness(self, factor: float) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        _check_adjust_brightness_errors_and_warnings(factor, plural=True)
        return _EmptyImageList()

    def add_noise(self, standard_deviation: float) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        _check_add_noise_errors(standard_deviation)
        return _EmptyImageList()

    def adjust_contrast(self, factor: float) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        _check_adjust_contrast_errors_and_warnings(factor, plural=True)
        return _EmptyImageList()

    def adjust_color_balance(self, factor: float) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        _check_adjust_color_balance_errors_and_warnings(factor, 0, plural=True)  # Disable channel check with channel=0
        return _EmptyImageList()

    def blur(self, radius: int) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        _check_blur_errors_and_warnings(radius, 1, plural=True)
        return _EmptyImageList()

    def sharpen(self, factor: float) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        _check_sharpen_errors_and_warnings(factor, plural=True)
        return _EmptyImageList()

    def invert_colors(self) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        return _EmptyImageList()

    def rotate_right(self) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        return _EmptyImageList()

    def rotate_left(self) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        return _EmptyImageList()

    def find_edges(self) -> ImageList:
        _EmptyImageList._warn_empty_image_list()
        return _EmptyImageList()



from __future__ import annotations

import io
import math
import os
import random
from abc import ABCMeta, abstractmethod
from pathlib import Path
from threading import Thread
from typing import TYPE_CHECKING, Literal, overload

from safeds._config import _init_default_device
from safeds._utils import _get_random_seed
from safeds._validation import _check_bounds, _ClosedBound
from safeds.data.image.containers._image import Image

if TYPE_CHECKING:
    from collections.abc import Sequence

    from torch import Tensor

    from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList
    from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
    from safeds.data.image.typing import ImageSize


class ImageList(metaclass=ABCMeta):
    """
    An ImageList is a list of different images. It can hold different sizes of Images. The channel of all images is the same.

    To create an `ImageList` call one of the following static methods:

    | Method                                                                        | Description                                              |
    | ----------------------------------------------------------------------------- | -------------------------------------------------------- |
    | [from_images][safeds.data.image.containers._image_list.ImageList.from_images] | Create an ImageList from a list of Images.               |
    | [from_files][safeds.data.image.containers._image_list.ImageList.from_files]   | Create an ImageList from a directory or a list of files. |
    """

    @staticmethod
    @abstractmethod
    def _create_image_list(images: list[Tensor], indices: list[int]) -> ImageList:
        """
        Create an ImageList from a list of tensors.

        Parameters
        ----------
        images:
            the list of tensors
        indices:
            a list of indices for the tensors

        Returns
        -------
        image_list:
            The image list
        """

    @staticmethod
    def from_images(images: list[Image]) -> ImageList:
        """
        Create an ImageList from a list of images.

        Parameters
        ----------
        images:
            the list of images

        Returns
        -------
        image_list:
            the image list
        """
        from safeds.data.image.containers._empty_image_list import _EmptyImageList
        from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList
        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        if len(images) == 0:
            return _EmptyImageList()

        indices = list(range(len(images)))
        first_width = images[0].width
        first_height = images[0].height
        for im in images:
            if first_width != im.width or first_height != im.height:
                return _MultiSizeImageList._create_image_list([image._image_tensor for image in images], indices)
        return _SingleSizeImageList._create_image_list([image._image_tensor for image in images], indices)

    @staticmethod
    @overload
    def from_files(path: str | Path | Sequence[str | Path]) -> ImageList: ...

    @staticmethod
    @overload
    def from_files(path: str | Path | Sequence[str | Path], *, load_percentage: float) -> ImageList: ...

    @staticmethod
    @overload
    def from_files(path: str | Path | Sequence[str | Path], *, return_filenames: Literal[False]) -> ImageList: ...

    @staticmethod
    @overload
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: Literal[False],
        load_percentage: float,
    ) -> ImageList: ...

    @staticmethod
    @overload
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: Literal[True],
    ) -> tuple[ImageList, list[str]]: ...

    @staticmethod
    @overload
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: Literal[True],
        load_percentage: float,
    ) -> tuple[ImageList, list[str]]: ...

    @staticmethod
    @overload
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: bool,
    ) -> ImageList | tuple[ImageList, list[str]]: ...

    @staticmethod
    @overload
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: bool,
        load_percentage: float,
    ) -> ImageList | tuple[ImageList, list[str]]: ...

    @staticmethod
    def from_files(
        path: str | Path | Sequence[str | Path],
        *,
        return_filenames: bool = False,
        load_percentage: float = 1.0,
    ) -> ImageList | tuple[ImageList, list[str]]:
        """
        Create an ImageList from a directory or a list of files.

        If you provide a path to a directory the images will be sorted alphabetically while inner directories will be sorted after image files.

        Parameters
        ----------
        path:
            the path to the directory or a list of files
        return_filenames:
            if True the output will be a tuple which contains a list of the filenames in order of the images
        load_percentage:
            the percentage of the given data being loaded. If below 1 the files will be shuffled before loading

        Returns
        -------
        image_list:
            the image list

        Raises
        ------
        FileNotFoundError
            If the directory or one of the files of the path cannot be found
        OutOfBoundsError
            If load_percentage is not between 0 and 1
        """
        from PIL.Image import open as pil_image_open

        _init_default_device()

        random.seed(_get_random_seed())

        from safeds.data.image.containers._empty_image_list import _EmptyImageList
        from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList
        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        _check_bounds("load_percentage", load_percentage, lower_bound=_ClosedBound(0), upper_bound=_ClosedBound(1))

        if isinstance(path, list) and len(path) == 0:
            return _EmptyImageList()

        file_names = []

        path_list: list[str | Path]
        if isinstance(path, Path | str):
            path_list = [Path(path)]
        else:
            path_list = list(path)
        while len(path_list) != 0:
            p = Path(path_list.pop(0))
            if p.is_dir():
                path_list += sorted([p / name for name in os.listdir(p)])
            elif p.is_file():
                file_names.append(str(p))
            else:
                raise FileNotFoundError(f"No such file or directory: '{path}'")

        if load_percentage < 1:
            random.shuffle(file_names)
            file_names = file_names[: max(round(len(file_names) * load_percentage), 1) if load_percentage > 0 else 0]

        num_of_files = len(file_names)

        if num_of_files == 0:
            return _EmptyImageList()

        image_sizes: dict[tuple[int, int], dict[int, list[str]]] = {}
        image_indices: dict[tuple[int, int], dict[int, list[int]]] = {}
        image_count: dict[tuple[int, int], int] = {}
        max_channel = -1

        for i, filename in enumerate(file_names):
            im = pil_image_open(filename)
            im_channel = len(im.getbands())
            im_size = (im.width, im.height)
            if im_channel > max_channel:
                max_channel = im_channel
            if im_size not in image_sizes:
                image_sizes[im_size] = {im_channel: [filename]}
                image_indices[im_size] = {im_channel: [i]}
                image_count[im_size] = 1
            elif im_channel not in image_sizes[im_size]:
                image_sizes[im_size][im_channel] = [filename]
                image_indices[im_size][im_channel] = [i]
                image_count[im_size] += 1
            else:
                image_sizes[im_size][im_channel].append(filename)
                image_indices[im_size][im_channel].append(i)
                image_count[im_size] += 1

        num_of_threads = min(math.ceil(num_of_files / 1000), 100)
        num_of_files_per_thread = math.ceil(num_of_files / num_of_threads)

        single_sized_image_lists = []
        thread_packages = []
        for size, image_files in image_sizes.items():
            im_list, packages = _SingleSizeImageList._create_image_list_from_files(
                image_files,
                image_count[size],
                max_channel,
                size[0],
                size[1],
                image_indices[size],
                num_of_files_per_thread,
            )
            single_sized_image_lists.append(im_list._as_single_size_image_list())
            thread_packages += packages
        thread_packages.sort(key=lambda x: len(x), reverse=True)

        threads: list[ImageList._FromImageThread] = []
        for thread_index in range(num_of_threads):
            current_thread_workload = 0
            current_thread_packages = []
            while current_thread_workload < num_of_files_per_thread and len(thread_packages) > 0:
                next_package = thread_packages.pop()
                current_thread_packages.append(next_package)
                current_thread_workload += len(next_package)
            if thread_index == num_of_threads - 1 and len(thread_packages) > 0:
                current_thread_packages += thread_packages  # pragma: no cover
            thread = ImageList._FromImageThread(current_thread_packages)
            threads.append(thread)
            thread.start()
        for thread in threads:
            thread.join()

        if len(single_sized_image_lists) == 1:
            image_list: ImageList = single_sized_image_lists[0]
        else:
            image_list = _MultiSizeImageList._create_from_single_sized_image_lists(single_sized_image_lists)

        if return_filenames:
            return image_list, file_names
        else:
            return image_list

    class _FromFileThreadPackage:
        def __init__(
            self,
            im_files: list[str],
            im_channel: int,
            to_channel: int,
            im_width: int,
            im_height: int,
            tensor: Tensor,
            start_index: int,
        ) -> None:
            self._im_files = im_files
            self._im_channel = im_channel
            self._to_channel = to_channel
            self._im_width = im_width
            self._im_height = im_height
            self._tensor = tensor
            self._start_index = start_index

        def load_files(self) -> None:
            import torch
            from torchvision.io import read_image

            _init_default_device()

            num_of_files = len(self._im_files)
            tensor_channel = max(self._im_channel, min(self._to_channel, 3))
            for index, im in enumerate(self._im_files):
                self._tensor[index + self._start_index, 0:tensor_channel] = read_image(im)
            if self._to_channel == 4 and self._im_channel < 4:
                torch.full(
                    (num_of_files, 1, self._im_height, self._im_width),
                    255,
                    out=self._tensor[self._start_index : self._start_index + num_of_files, 3:4],
                )

        def __len__(self) -> int:
            return len(self._im_files)

    class _FromImageThread(Thread):
        def __init__(self, packages: list[ImageList._FromFileThreadPackage]) -> None:
            super().__init__()
            self._packages = packages

        def run(self) -> None:
            for pck in self._packages:
                pck.load_files()

    @abstractmethod
    def _clone(self) -> ImageList:
        """
        Clone your ImageList to a new instance.

        Returns
        -------
        image_list:
            the cloned image list
        """

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """
        Compare two image lists.

        Two image lists are only equal if they contain the same images at the same indices.

        Parameters
        ----------
        other:
            the image list to compare to

        Returns
        -------
        equals:
            Whether the two image lists are equal
        """

    @abstractmethod
    def __hash__(self) -> int:
        """
        Return a deterministic hash value for this image list.

        Returns
        -------
        hash:
            The hash value.
        """

    @abstractmethod
    def __sizeof__(self) -> int:
        """
        Return the complete size of this object.

        Returns
        -------
        size:
            Size of this object in bytes.
        """

    def __len__(self) -> int:
        """
        Return the number of images in this image list.

        Returns
        -------
        length:
            The number of images
        """
        return self.image_count

    def __contains__(self, item: object) -> bool:
        """
        Return whether the given item is in this image list.

        Parameters
        ----------
        item:
            the item to check

        Returns
        -------
        has_item:
            Whether the given item is in this image list
        """
        return isinstance(item, Image) and self.has_image(item)

    def _repr_png_(self) -> bytes:
        """
        Return a PNG representation of this image list as bytes.

        Returns
        -------
        png:
            the png representation of this image list
        """
        import torch
        from torchvision.utils import make_grid, save_image

        _init_default_device()

        from safeds.data.image.containers._empty_image_list import _EmptyImageList

        if isinstance(self, _EmptyImageList):
            raise TypeError("You cannot display an empty ImageList")

        max_width, max_height = max(self.widths), max(self.heights)
        tensors = []
        for image in self.to_images():
            im_tensor = torch.zeros([4, max_height, max_width])
            im_tensor[:, : image.height, : image.width] = image.change_channel(4)._image_tensor
            tensors.append(im_tensor)
        tensor_grid = make_grid(tensors, math.ceil(math.sqrt(len(tensors))))
        buffer = io.BytesIO()
        save_image(tensor_grid.to(torch.float32) / 255, buffer, format="png")
        buffer.seek(0)
        return buffer.read()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def image_count(self) -> int:
        """The number of images in this image list."""

    @property
    @abstractmethod
    def widths(self) -> list[int]:
        """A list of all widths in this image list."""

    @property
    @abstractmethod
    def heights(self) -> list[int]:
        """A list of all heights in this image list."""

    @property
    @abstractmethod
    def channel(self) -> int:
        """The channel of all images."""

    @property
    @abstractmethod
    def sizes(self) -> list[ImageSize]:
        """The sizes of all images."""

    @property
    @abstractmethod
    def size_count(self) -> int:
        """The number of different sizes of images in this image list."""

    # ------------------------------------------------------------------------------------------------------------------
    # Getters
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def get_image(self, index: int) -> Image:
        """
        Return the image at the given index.

        Parameters
        ----------
        index:
            the index for the image to return

        Returns
        -------
        image:
            the image at the given index
        """

    @abstractmethod
    def index(self, image: Image) -> list[int]:
        """
        Return a list of indexes of the given image.

        If the image has multiple occurrences, all indices will be returned

        Parameters
        ----------
        image:
            the image to search for occurrences

        Returns
        -------
        indices:
            all occurrences of the image
        """

    @abstractmethod
    def has_image(self, image: Image) -> bool:
        """
        Return whether the given image is in this image list.

        Parameters
        ----------
        image:
            the image to check

        Returns
        -------
        has_image:
            Whether the given image is in this image list
        """

    # ------------------------------------------------------------------------------------------------------------------
    # Conversion
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def to_jpeg_files(self, path: str | Path | list[str | Path]) -> None:
        """
        Save all images as jpeg files.

        Parameters
        ----------
        path:
            Either the path to a directory or a list of directories which has directories for either all different sizes or all different images. Any non-existant path will be created

        Raises
        ------
        IllegalFormatError
            If the channel of the images is not supported
        ValueError
            If the path is a list but has too few or too many entries
        """

    @abstractmethod
    def to_png_files(self, path: str | Path | list[str | Path]) -> None:
        """
        Save all images as png files.

        Parameters
        ----------
        path:
            Either the path to a directory or a list of directories which has directories for either all different sizes or all different images. Any non-existant path will be created

        Raises
        ------
        ValueError
            If the path is a list but has too few or too many entries
        """

    @abstractmethod
    def to_images(self, indices: list[int] | None = None) -> list[Image]:
        """
        Return a list of all images in this image list.

        Parameters
        ----------
        indices:
            a list of all indices to include in the output. If None, all indices will be included

        Returns
        -------
        images:
            the list of all images

        Raises
        ------
        IndexOutOfBoundsError
            If any index is out of bounds
        """

    def _as_multi_size_image_list(self) -> _MultiSizeImageList:
        """
        Typechecking method for MultiSizeImageList.

        Returns
        -------
        self:
            self as a MultiSizeImageList

        Raises
        ------
        ValueError
            if this image list is not a MultiSizeImageList
        """
        from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList

        if isinstance(self, _MultiSizeImageList):
            return self
        raise ValueError("The given image_list is not a MultiSizeImageList")

    def _as_single_size_image_list(self) -> _SingleSizeImageList:
        """
        Typechecking method for SingleSizeImageList.

        Returns
        -------
        self:
            self as a SingleSizeImageList

        Raises
        ------
        ValueError
            if this image list is not a SingleSizeImageList
        """
        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        if isinstance(self, _SingleSizeImageList):
            return self
        raise ValueError("The given image_list is not a SingleSizeImageList")

    # ------------------------------------------------------------------------------------------------------------------
    # Transformations
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def change_channel(self, channel: int) -> ImageList:
        """
        Return a new `ImageList` that has the given number of channels.

        The original image list is not modified.

        Parameters
        ----------
        channel:
            The new number of channels. 1 will result in grayscale images

        Returns
        -------
        image_list:
            the image list with the given number of channels

        Raises
        ------
        ValueError:
            if the given channel is not a valid channel option
        """

    @abstractmethod
    def _add_image_tensor(self, image_tensor: Tensor, index: int) -> ImageList:
        """
        Return a new `ImageList` with the given tensor added as an image.

        The original image list is not modified.

        Parameters
        ----------
        image_tensor:
            The new tensor to be added to the image list

        Returns
        -------
        image_list:
            the image list with the new tensor added
        """

    def add_image(self, image: Image) -> ImageList:
        """
        Return a new `ImageList` with the given image added to the image list.

        The original image list is not modified.

        Parameters
        ----------
        image:
            The image to be added to the image list

        Returns
        -------
        image_list:
            the image list with the new image added
        """
        return self._add_image_tensor(image._image_tensor, self.image_count)

    @abstractmethod
    def add_images(self, images: list[Image] | ImageList) -> ImageList:
        """
        Return a new `ImageList` with the given images added to the image list.

        The original image list is not modified.

        Parameters
        ----------
        images:
            The images to be added to the image list

        Returns
        -------
        image_list:
            the image list with the new images added
        """

    def remove_image(self, image: Image) -> ImageList:
        """
        Return a new `ImageList` with the given image removed from the image list.

        If the image has multiple occurrences, all occurrences will be removed.

        The original image list is not modified.

        Parameters
        ----------
        image:
            The image to be removed from the image list

        Returns
        -------
        image_list:
            the image list with the given image removed
        """
        return self._remove_image_by_index_ignore_invalid(self.index(image))

    def remove_images(self, images: list[Image]) -> ImageList:
        """
        Return a new `ImageList` with the given images removed from the image list.

        If one image has multiple occurrences, all occurrences will be removed.

        The original image list is not modified.

        Parameters
        ----------
        images:
            The images to be removed from the image list

        Returns
        -------
        image_list:
            the image list with the given images removed
        """
        indices_to_remove = []
        for image in images:
            indices_to_remove += self.index(image)
        return self._remove_image_by_index_ignore_invalid(list(set(indices_to_remove)))

    @abstractmethod
    def remove_image_by_index(self, index: int | list[int]) -> ImageList:
        """
        Return a new `ImageList` with the given indices removed from the image list.

        The original image list is not modified.

        Parameters
        ----------
        index:
            The index of the image to be removed from the image list

        Returns
        -------
        image_list:
            the image list with the without the removed image

        Raises
        ------
        IndexOutOfBoundsError
            If one of the indices is out of bounds
        """

    @abstractmethod
    def _remove_image_by_index_ignore_invalid(self, index: int | list[int]) -> ImageList:
        """
        Return a new `ImageList` with the given indices removed from the image list.

        Invalid indices will be ignored.

        The original image list is not modified.

        Parameters
        ----------
        index:
            The index of the image to be removed from the image list

        Returns
        -------
        image_list:
            the image list with the without the removed image
        """

    @abstractmethod
    def remove_images_with_size(self, width: int, height: int) -> ImageList:
        """
        Return a new `ImageList` with the all images of the given size removed.

        The original image list is not modified.

        Parameters
        ----------
        width:
            The width of the images to be removed from the image list
        height:
            The height of the images to be removed from the image list

        Returns
        -------
        image_list:
            the image list with the given images removed

        Raises
        ------
        OutOfBoundsError
            If width or height are below 1
        """

    @abstractmethod
    def remove_duplicate_images(self) -> ImageList:
        """
        Return a new `ImageList` with all duplicate images removed.

        One occurrence of each image will stay in the image list.

        The original image list is not modified.

        Returns
        -------
        image_list:
            the image list with only unique images
        """

    @abstractmethod
    def shuffle_images(self) -> ImageList:
        """
        Return a new `ImageList` with all images shuffled.

        The original image list is not modified.

        Returns
        -------
        image_list:
            the image list with shuffled images
        """

    @abstractmethod
    def resize(self, new_width: int, new_height: int) -> ImageList:
        """
        Return a new `ImageList` with all images resized to a given size.

        The original image list is not modified.

        Parameters
        ----------
        new_width:
            the new width of the images
        new_height:
            the new height of the images

        Returns
        -------
        image_list:
            The image list with all images resized to the given width and height.

        Raises
        ------
        OutOfBoundsError
            If new_width or new_height are below 1
        """

    @abstractmethod
    def convert_to_grayscale(self) -> ImageList:
        """
        Return a new `ImageList` with all images converted to grayscale.

        The new image list will have the same amount of channels as the original image list.
        If you want to change the amount of channels used, please use the method [change_channel][safeds.data.image.containers._image_list.ImageList.change_channel].

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with all images converted to grayscale.
        """

    @abstractmethod
    def crop(self, x: int, y: int, width: int, height: int) -> ImageList:
        """
        Return a new `ImageList` with all images cropped to a given bounding rectangle.

        The original image list is not modified.

        Parameters
        ----------
        x:
            the x coordinate of the top-left corner of the bounding rectangle
        y:
            the y coordinate of the top-left corner of the bounding rectangle
        width:
            the width of the bounding rectangle
        height:
            the height of the bounding rectangle

        Returns
        -------
        image_list:
            The image list with all images cropped

        Raises
        ------
        OutOfBoundsError
            If x or y are below 0 or if width or height are below 1
        """

    @abstractmethod
    def flip_vertically(self) -> ImageList:
        """
        Return a new `ImageList` with all images flipped vertically (horizontal axis, flips up-down and vice versa).

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with all images flipped vertically
        """

    @abstractmethod
    def flip_horizontally(self) -> ImageList:
        """
        Return a new `ImageList` with all images flipped horizontally (vertical axis, flips left-right and vice versa).

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with all images flipped horizontally
        """

    @abstractmethod
    def adjust_brightness(self, factor: float) -> ImageList:
        """
        Return a new `ImageList` where all images have the adjusted brightness.

        The original image list is not modified.

        Parameters
        ----------
        factor:
            The brightness factor.
            1.0 will not change the brightness.
            Below 1.0 will result in a darker images.
            Above 1.0 will resolut in a brighter images.
            Has to be bigger than or equal to 0 (black).

        Returns
        -------
        image_list:
            The image list with adjusted brightness

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """

    @abstractmethod
    def add_noise(self, standard_deviation: float) -> ImageList:
        """
        Return a new `ImageList` with noise added to all images.

        The original image list is not modified.

        Parameters
        ----------
        standard_deviation:
            The standard deviation of the normal distribution. Has to be bigger than or equal to 0.

        Returns
        -------
        image_list:
            The image list with added noise

        Raises
        ------
        OutOfBoundsError
            If standard_deviation is smaller than 0.
        """

    @abstractmethod
    def adjust_contrast(self, factor: float) -> ImageList:
        """
        Return a new `ImageList` where all images have the adjusted contrast.

        The original image list is not modified.

        Parameters
        ----------
        factor:
            If factor > 1, increase contrast of images.
            If factor = 1, no changes will be made.
            If factor < 1, make images greyer.
            Has to be bigger than or equal to 0 (gray).

        Returns
        -------
        image_list:
            The image list with adjusted contrast

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """

    @abstractmethod
    def adjust_color_balance(self, factor: float) -> ImageList:
        """
        Return a new `ImageList` where all images have the adjusted color balance.

        The original image list is not modified.

        Parameters
        ----------
        factor:
            Has to be bigger than or equal to 0.
            If 0 <= factor < 1, make images greyer.
            If factor = 1, no changes will be made.
            If factor > 1, increase color balance of images.

        Returns
        -------
        image_list:
            The image list with adjusted color balance

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """

    @abstractmethod
    def blur(self, radius: int) -> ImageList:
        """
        Return a new `ImageList` where all images have been blurred.

        The original image list is not modified.

        Parameters
        ----------
        radius:
             Radius is directly proportional to the blur value. The radius is equal to the amount of pixels united in
             each direction. A radius of 1 will result in a united box of 9 pixels.

        Returns
        -------
        image_list:
            The image list with blurred images

        Raises
        ------
        OutOfBoundsError
            If radius is smaller than 0 or equal or greater than the smallest size of one of the images.
        """

    @abstractmethod
    def sharpen(self, factor: float) -> ImageList:
        """
        Return a new `ImageList` where all images have been sharpened.

        The original image list is not modified.

        Parameters
        ----------
        factor:
            If factor > 1, increase the sharpness of the images.
            If factor = 1, no changes will be made.
            If factor < 1, blur the images.
            Has to be bigger than or equal to 0 (blurred).

        Returns
        -------
        image_list:
            The image list with sharpened images

        Raises
        ------
        OutOfBoundsError
            If factor is smaller than 0.
        """

    @abstractmethod
    def invert_colors(self) -> ImageList:
        """
        Return a new `ImageList` where all images have their colors inverted.

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with inverted colors
        """

    @abstractmethod
    def rotate_right(self) -> ImageList:
        """
        Return a new `ImageList` where all images have been rotated 90 degrees clockwise.

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with all images rotated
        """

    @abstractmethod
    def rotate_left(self) -> ImageList:
        """
        Return a new `ImageList` where all images have been rotated 90 degrees counter-clockwise.

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with all images rotated
        """

    @abstractmethod
    def find_edges(self) -> ImageList:
        """
        Return a new `ImageList` with grayscale versions of the images with the edges highlighted.

        The original image list is not modified.

        Returns
        -------
        image_list:
            The image list with highlighted edges
        """



"""Types used to define the attributes of image data."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._image_size import ImageSize

apipkg.initpkg(
    __name__,
    {
        "ImageSize": "._image_size:ImageSize",
    },
)

__all__ = [
    "ImageSize",
]



from __future__ import annotations


class NonNumericColumnError(TypeError):
    """Exception raised for trying to do numerical operations on a non-numerical column."""

    def __init__(self, column_info: str, help_msg: str | None = None) -> None:
        line_break = "\n"
        super().__init__(
            "Tried to do a numerical operation on one or multiple non-numerical columns:"
            f" \n{column_info}{line_break + help_msg if help_msg is not None else ''}",
        )


class MissingValuesColumnError(Exception):
    """Exception raised for trying to do operations on columns containing missing values."""

    def __init__(self, column_info: str, help_msg: str | None = None) -> None:
        line_break = "\n"
        super().__init__(
            "Tried to do an operation on one or multiple columns containing missing values:"
            f" \n{column_info}{line_break + help_msg if help_msg is not None else ''}",
        )


class DuplicateColumnError(ValueError):
    """
    Exception raised for trying to modify a table resulting in a duplicate column name.

    Parameters
    ----------
    column_name:
        The name of the column that resulted in a duplicate.
    """

    def __init__(self, column_name: str):
        super().__init__(f"Column '{column_name}' already exists.")


class IndexOutOfBoundsError(IndexError):
    """
    Exception raised for trying to access an element by an index that does not exist in the underlying data.

    Parameters
    ----------
    index:
        The wrongly used index.
    """

    def __init__(self, index: int | list[int] | slice):
        if isinstance(index, list):
            if len(index) == 1:
                index = index[0]
            else:
                super().__init__(f"There are no elements at indices {index}.")
                return
        if isinstance(index, int):
            super().__init__(f"There is no element at index '{index}'.")
        else:
            super().__init__(f"There is no element in the range [{index.start}, {index.stop}].")


class DuplicateIndexError(IndexError):
    """
    Exception raised for trying to add an element with an index that does already exist in the underlying data.

    Parameters
    ----------
    index:
        The wrongly added index.
    """

    def __init__(self, index: int):
        super().__init__(f"The index '{index}' is already in use.")


class ColumnSizeError(Exception):
    """
    Exception raised for trying to use a column of unsupported size.

    Parameters
    ----------
    expected_size:
        The expected size of the column as an expression (e.g. 2, >0, !=0).
    actual_size:
        The actual size of the column as an expression (e.g. 2, >0, !=0).
    """

    def __init__(self, expected_size: str, actual_size: str):
        super().__init__(f"Expected a column of size {expected_size} but got column of size {actual_size}.")


class ColumnLengthMismatchError(ValueError):
    """Exception raised when the lengths of two or more columns do not match."""

    def __init__(self, column_info: str):
        super().__init__(f"The length of at least one column differs: \n{column_info}")


class OutputLengthMismatchError(Exception):
    """Exception raised when the lengths of the input and output container does not match."""

    def __init__(self, output_info: str):
        super().__init__(f"The length of the output container differs: \n{output_info}")


class TransformerNotFittedError(Exception):
    """Raised when a transformer is used before fitting it."""

    def __init__(self) -> None:
        super().__init__("The transformer has not been fitted yet.")


class TransformerNotInvertibleError(Exception):
    """Raised when a function tries to invert a non-invertible transformer."""

    def __init__(self, transformer_type: str) -> None:
        super().__init__(f"{transformer_type} is not invertible.")


class ValueNotPresentWhenFittedError(Exception):
    """Exception raised when attempting to one-hot-encode a table containing values not present in the fitting phase."""

    def __init__(self, values: list[tuple[str, str]]) -> None:
        values_info = [f"{value} in column {column}" for value, column in values]
        line_break = "\n"
        super().__init__(
            "Value(s) not present in the table the transformer was fitted on:"
            f" {line_break}{line_break.join(values_info)}",
        )


class IllegalFormatError(Exception):
    """Exception raised when a format is not legal."""

    def __init__(self, formats: list[str] | str) -> None:
        super().__init__(f"This format is illegal. Use one of the following formats: {formats}")



"""Types used to define neural networks and related attributes."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._model_image_size import ConstantImageSize, ModelImageSize, VariableImageSize

apipkg.initpkg(
    __name__,
    {
        "ConstantImageSize": "._model_image_size:ConstantImageSize",
        "ModelImageSize": "._model_image_size:ModelImageSize",
        "VariableImageSize": "._model_image_size:VariableImageSize",
    },
)

__all__ = [
    "ConstantImageSize",
    "ModelImageSize",
    "VariableImageSize",
]



from __future__ import annotations

from abc import ABC, abstractmethod


def _check_bounds(
    name: str,
    actual: float | None,
    *,
    lower_bound: _Bound | None = None,
    upper_bound: _Bound | None = None,
) -> None:
    """
    Check that a value is within the expected range and raise an error if it is not.

    Parameters
    ----------
    actual:
        The actual value that should be checked.
    name:
        The name of the offending variable.
    lower_bound:
        The lower bound of the expected range. Use None if there is no lower Bound.
    upper_bound:
        The upper bound of the expected range. Use None if there is no upper Bound.

    Raises
    ------
    OutOfBoundsError:
        If the actual value is outside its expected range.
    """
    from safeds.exceptions import OutOfBoundsError  # circular import

    if actual is None:
        return  # Skip the check if the actual value is None (i.e., not provided).

    if lower_bound is None:
        lower_bound = _OpenBound(float("-inf"))
    if upper_bound is None:
        upper_bound = _OpenBound(float("inf"))

    if not lower_bound._check_as_lower_bound(actual) or not upper_bound._check_as_upper_bound(actual):
        message = _build_error_message(name, actual, lower_bound, upper_bound)
        raise OutOfBoundsError(message)


def _build_error_message(name: str, actual: float, lower_bound: _Bound, upper_bound: _Bound) -> str:
    range_ = f"{lower_bound._to_string_as_lower_bound()}, {upper_bound._to_string_as_upper_bound()}"
    return f"{name} must be in {range_} but was {actual}."


class _Bound(ABC):
    """Lower or upper bound of the legal range of a value."""

    @abstractmethod
    def _check_as_lower_bound(self, actual: float) -> bool:
        """
        Treat this bound as the lower bound and check that a value does not exceed it.

        Parameters
        ----------
        actual:
            The actual value to check.

        Returns
        -------
        in_bounds:
            Whether the actual value is within the expected range.
        """

    @abstractmethod
    def _check_as_upper_bound(self, actual: float) -> bool:
        """
        Treat this bound as the upper bound and check that a value does not exceed it.

        Parameters
        ----------
        actual:
            The actual value to check.

        Returns
        -------
        in_bounds:
            Whether the actual value is within the expected range.
        """

    @abstractmethod
    def _to_string_as_lower_bound(self) -> str:
        """Treat this bound as the lower bound and get a string representation."""

    @abstractmethod
    def _to_string_as_upper_bound(self) -> str:
        """Treat this bound as the upper bound and get a string representation."""


class _ClosedBound(_Bound):
    """
    A closed bound where the border value belongs to the range.

    Parameters
    ----------
    value:
        The border value of the bound.
    """

    def __init__(self, value: float):
        self.value: float = value

    def _check_as_lower_bound(self, actual: float) -> bool:
        return actual >= self.value

    def _check_as_upper_bound(self, actual: float) -> bool:
        return actual <= self.value

    def _to_string_as_lower_bound(self) -> str:
        return f"[{_float_to_string(self.value)}"

    def _to_string_as_upper_bound(self) -> str:
        return f"{_float_to_string(self.value)}]"


class _OpenBound(_Bound):
    """
    An open bound where the border value does not belong to the range.

    Parameters
    ----------
    value:
        The border value of the bound.
    """

    def __init__(self, value: float):
        self.value: float = value

    def _check_as_lower_bound(self, actual: float) -> bool:
        return actual > self.value

    def _check_as_upper_bound(self, actual: float) -> bool:
        return actual < self.value

    def _to_string_as_lower_bound(self) -> str:
        return f"({_float_to_string(self.value)}"

    def _to_string_as_upper_bound(self) -> str:
        return f"{_float_to_string(self.value)})"


def _float_to_string(value: float) -> str:
    if value == float("-inf"):
        return "-\u221e"
    elif value == float("inf"):
        return "\u221e"
    else:
        return str(value)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING
from warnings import warn

from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _ClosedBound

from ._regressor import Regressor

if TYPE_CHECKING:
    from sklearn.base import RegressorMixin


class ElasticNetRegressor(Regressor):
    """Elastic net regression.

    Parameters
    ----------
    alpha:
        Controls the regularization of the model. The higher the value, the more regularized it becomes.
    lasso_ratio:
        Number between 0 and 1 that controls the ratio between Lasso and Ridge regularization. If 0, only Ridge
        regularization is used. If 1, only Lasso regularization is used.

    Raises
    ------
    OutOfBoundsError
        If `alpha` is negative or `lasso_ratio` is not between 0 and 1.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, *, alpha: float = 1.0, lasso_ratio: float = 0.5) -> None:
        super().__init__()

        # Validation
        _check_bounds("alpha", alpha, lower_bound=_ClosedBound(0))
        if alpha == 0:
            warn(
                (
                    "Setting alpha to zero makes this model equivalent to LinearRegression. You should use "
                    "LinearRegression instead for better numerical stability."
                ),
                UserWarning,
                stacklevel=2,
            )

        _check_bounds("lasso_ratio", lasso_ratio, lower_bound=_ClosedBound(0), upper_bound=_ClosedBound(1))
        if lasso_ratio == 0:
            warnings.warn(
                (
                    "ElasticNetRegression with lasso_ratio = 0 is essentially RidgeRegression."
                    " Use RidgeRegression instead for better numerical stability."
                ),
                stacklevel=2,
            )
        elif lasso_ratio == 1:
            warnings.warn(
                (
                    "ElasticNetRegression with lasso_ratio = 0 is essentially LassoRegression."
                    " Use LassoRegression instead for better numerical stability."
                ),
                stacklevel=2,
            )

        # Hyperparameters
        self._alpha = alpha
        self._lasso_ratio = lasso_ratio

    def __hash__(self) -> int:
        return _structural_hash(
            super().__hash__(),
            self._alpha,
            self._lasso_ratio,
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def alpha(self) -> float:
        """The regularization of the model."""
        return self._alpha

    @property
    def lasso_ratio(self) -> float:
        """Rhe ratio between Lasso and Ridge regularization."""
        return self._lasso_ratio

    # ------------------------------------------------------------------------------------------------------------------
    # Template methods
    # ------------------------------------------------------------------------------------------------------------------

    def _clone(self) -> ElasticNetRegressor:
        return ElasticNetRegressor(
            alpha=self._alpha,
            lasso_ratio=self._lasso_ratio,
        )

    def _get_sklearn_model(self) -> RegressorMixin:
        from sklearn.linear_model import ElasticNet as SklearnElasticNet

        return SklearnElasticNet(
            alpha=self._alpha,
            l1_ratio=self._lasso_ratio,
        )



from __future__ import annotations

import copy
import random
import sys
from typing import TYPE_CHECKING

from safeds._config import _init_default_device
from safeds._utils import _structural_hash
from safeds.data.image._utils._image_transformation_error_and_warning_checks import (
    _check_blur_errors_and_warnings,
    _check_crop_errors,
    _check_remove_images_with_size_errors,
    _check_resize_errors,
)
from safeds.data.image.containers import Image, ImageList
from safeds.exceptions import (
    DuplicateIndexError,
    IllegalFormatError,
    IndexOutOfBoundsError,
)

if TYPE_CHECKING:
    from pathlib import Path

    from torch import Tensor

    from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
    from safeds.data.image.typing import ImageSize


class _MultiSizeImageList(ImageList):
    """
    An ImageList is a list of different images. It can hold different sizes of Images. The channel of all images is the same.

    This is the class for an ImageList with multiple different sizes.

    To create an `ImageList` call one of the following static methods:

    | Method                                                                        | Description                                              |
    | ----------------------------------------------------------------------------- | -------------------------------------------------------- |
    | [from_images][safeds.data.image.containers._image_list.ImageList.from_images] | Create an ImageList from a list of Images.               |
    | [from_files][safeds.data.image.containers._image_list.ImageList.from_files]   | Create an ImageList from a directory or a list of files. |
    """

    def __init__(self) -> None:
        self._image_list_dict: dict[tuple[int, int], ImageList] = {}  # {image_size: image_list}
        self._indices_to_image_size_dict: dict[int, tuple[int, int]] = {}  # {index: image_size}

    @staticmethod
    def _create_from_single_sized_image_lists(single_size_image_lists: list[_SingleSizeImageList]) -> ImageList:
        from safeds.data.image.containers._empty_image_list import _EmptyImageList

        if len(single_size_image_lists) == 0:
            return _EmptyImageList()
        elif len(single_size_image_lists) == 1:
            return single_size_image_lists[0]

        different_channels: bool = False
        max_channel: None | int = None

        image_list = _MultiSizeImageList()
        for single_size_image_list in single_size_image_lists:
            image_size = (single_size_image_list.widths[0], single_size_image_list.heights[0])
            image_list._image_list_dict[image_size] = single_size_image_list
            image_list._indices_to_image_size_dict.update(
                zip(
                    single_size_image_list._indices_to_tensor_positions.keys(),
                    [image_size] * len(single_size_image_list),
                    strict=False,
                ),
            )
            if max_channel is None:
                max_channel = single_size_image_list.channel
            elif max_channel < single_size_image_list.channel:
                different_channels = True
                max_channel = single_size_image_list.channel
            elif max_channel > single_size_image_list.channel:
                different_channels = True

        if different_channels:
            for size in image_list._image_list_dict:
                if max_channel is not None and image_list._image_list_dict[size].channel != max_channel:
                    image_list._image_list_dict[size] = image_list._image_list_dict[size].change_channel(
                        int(max_channel),
                    )
        return image_list

    @staticmethod
    def _create_image_list(images: list[Tensor], indices: list[int]) -> ImageList:
        from safeds.data.image.containers._empty_image_list import _EmptyImageList
        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        if len(images) == 0:
            return _EmptyImageList()

        max_channel = 0

        image_tensor_dict = {}
        image_index_dict = {}

        image_list = _MultiSizeImageList()
        for index in indices:
            image = images.pop(0)
            size = (image.size(dim=2), image.size(dim=1))
            if size not in image_tensor_dict:
                image_tensor_dict[size] = [image]
                image_index_dict[size] = [index]
                max_channel = max(max_channel, image.size(dim=-3))
            else:
                image_tensor_dict[size].append(image)
                image_index_dict[size].append(index)
                max_channel = max(max_channel, image.size(dim=-3))

        for size in image_tensor_dict:
            image_list._image_list_dict[size] = _SingleSizeImageList._create_image_list(
                image_tensor_dict[size],
                image_index_dict[size],
            )
            image_list._indices_to_image_size_dict.update(
                zip(
                    image_list._image_list_dict[size]._as_single_size_image_list()._indices_to_tensor_positions.keys(),
                    [size] * len(image_list._image_list_dict[size]),
                    strict=False,
                ),
            )

        if max_channel > 1:
            image_list = image_list.change_channel(max_channel)._as_multi_size_image_list()

        return image_list

    def _clone(self) -> ImageList:
        cloned_image_list = self._clone_without_image_dict()
        for image_list_size, image_list in self._image_list_dict.items():
            cloned_image_list._image_list_dict[image_list_size] = image_list._clone()
        return cloned_image_list

    def _clone_without_image_dict(self) -> _MultiSizeImageList:
        """
        Clone this MultiSizeImageList to a new instance without the image data.

        Returns
        -------
        image_list:
            the cloned image list
        """
        cloned_image_list = _MultiSizeImageList()
        cloned_image_list._indices_to_image_size_dict = copy.deepcopy(self._indices_to_image_size_dict)
        return cloned_image_list

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ImageList):
            return NotImplemented
        if not isinstance(other, _MultiSizeImageList) or set(other._image_list_dict) != set(self._image_list_dict):
            return False
        if self is other:
            return True
        for image_list_key, image_list_value in self._image_list_dict.items():
            if image_list_value != other._image_list_dict[image_list_key]:
                return False
        return True

    def __hash__(self) -> int:
        return _structural_hash([self._image_list_dict[image_size] for image_size in sorted(self._image_list_dict)])

    def __sizeof__(self) -> int:
        return (
            sys.getsizeof(self._image_list_dict)
            + sum(map(sys.getsizeof, self._image_list_dict.keys()))
            + sum(map(sys.getsizeof, self._image_list_dict.values()))
            + sys.getsizeof(self._indices_to_image_size_dict)
            + sum(map(sys.getsizeof, self._indices_to_image_size_dict.keys()))
            + sum(map(sys.getsizeof, self._indices_to_image_size_dict.values()))
        )

    @property
    def image_count(self) -> int:
        length = 0
        for image_list in self._image_list_dict.values():
            length += len(image_list)
        return length

    @property
    def widths(self) -> list[int]:
        widths = {}
        for image_list in self._image_list_dict.values():
            indices = image_list._as_single_size_image_list()._tensor_positions_to_indices
            for i, index in enumerate(indices):
                widths[index] = image_list.widths[i]
        return [widths[index] for index in sorted(widths)]

    @property
    def heights(self) -> list[int]:
        heights = {}
        for image_list in self._image_list_dict.values():
            indices = image_list._as_single_size_image_list()._tensor_positions_to_indices
            for i, index in enumerate(indices):
                heights[index] = image_list.heights[i]
        return [heights[index] for index in sorted(heights)]

    @property
    def channel(self) -> int:
        return next(iter(self._image_list_dict.values())).channel

    @property
    def sizes(self) -> list[ImageSize]:
        sizes = {}
        for image_list in self._image_list_dict.values():
            indices = image_list._as_single_size_image_list()._tensor_positions_to_indices
            for i, index in enumerate(indices):
                sizes[index] = image_list.sizes[i]
        return [sizes[index] for index in sorted(sizes)]

    @property
    def size_count(self) -> int:
        return len(self._image_list_dict)

    def get_image(self, index: int) -> Image:
        if index not in self._indices_to_image_size_dict:
            raise IndexOutOfBoundsError(index)
        return self._image_list_dict[self._indices_to_image_size_dict[index]].get_image(index)

    def index(self, image: Image) -> list[int]:
        indices = []
        for image_list in self._image_list_dict.values():
            if image in image_list:
                indices += image_list.index(image)
        return indices

    def has_image(self, image: Image) -> bool:
        return (image.width, image.height) in self._image_list_dict and self._image_list_dict[
            (image.width, image.height)
        ].has_image(image)

    def to_jpeg_files(self, path: str | Path | list[str | Path]) -> None:
        if self.channel == 4:
            raise IllegalFormatError("png")
        if isinstance(path, list):
            if len(path) == self.image_count:
                for image_size, image_list in self._image_list_dict.items():
                    image_list.to_jpeg_files(
                        [p for i, p in enumerate(path) if self._indices_to_image_size_dict[i] == image_size],
                    )
            elif len(path) == self.size_count:
                image_list_path: str | Path
                for image_list_path, image_list in zip(path, self._image_list_dict.values(), strict=False):
                    image_list.to_jpeg_files(image_list_path)
            else:
                raise ValueError(
                    "The path specified is invalid. Please provide either the path to a directory, a list of paths with one path for each image, or a list of paths with one path per image size.",
                )
        else:
            for image_list in self._image_list_dict.values():
                image_list.to_jpeg_files(path)

    def to_png_files(self, path: str | Path | list[str | Path]) -> None:
        if isinstance(path, list):
            if len(path) == self.image_count:
                for image_size, image_list in self._image_list_dict.items():
                    image_list.to_png_files(
                        [p for i, p in enumerate(path) if self._indices_to_image_size_dict[i] == image_size],
                    )
            elif len(path) == self.size_count:
                image_list_path: str | Path
                for image_list_path, image_list in zip(path, self._image_list_dict.values(), strict=False):
                    image_list.to_png_files(image_list_path)
            else:
                raise ValueError(
                    "The path specified is invalid. Please provide either the path to a directory, a list of paths with one path for each image, or a list of paths with one path per image size.",
                )
        else:
            for image_list in self._image_list_dict.values():
                image_list.to_png_files(path)

    def to_images(self, indices: list[int] | None = None) -> list[Image]:
        if indices is None:
            indices = sorted(self._indices_to_image_size_dict)
        else:
            wrong_indices = []
            for index in indices:
                if index not in self._indices_to_image_size_dict:
                    wrong_indices.append(index)
            if len(wrong_indices) != 0:
                raise IndexOutOfBoundsError(wrong_indices)
        images = []
        for index in indices:
            images.append(self._image_list_dict[self._indices_to_image_size_dict[index]].get_image(index))
        return images

    def change_channel(self, channel: int) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.change_channel(channel)
        return image_list

    def _add_image_tensor(self, image_tensor: Tensor, index: int) -> ImageList:
        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        if index in self._indices_to_image_size_dict:
            raise DuplicateIndexError(index)

        image_list = self._clone()._as_multi_size_image_list()
        size = (image_tensor.size(dim=2), image_tensor.size(dim=1))
        image_list._indices_to_image_size_dict[index] = size

        if size in self._image_list_dict:
            image_list._image_list_dict[size] = image_list._image_list_dict[size]._add_image_tensor(image_tensor, index)
        else:
            image_list._image_list_dict[size] = _SingleSizeImageList._create_image_list([image_tensor], [index])

        if image_tensor.size(dim=0) != self.channel:
            image_list = image_list.change_channel(
                max(image_tensor.size(dim=0), self.channel),
            )._as_multi_size_image_list()

        return image_list

    def add_images(self, images: list[Image] | ImageList) -> ImageList:
        from safeds.data.image.containers._empty_image_list import _EmptyImageList
        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        if isinstance(images, _EmptyImageList) or isinstance(images, list) and len(images) == 0:
            return self

        indices_for_images_with_size = {}
        current_index = max(self._indices_to_image_size_dict) + 1
        image_list_with_size: dict[tuple[int, int], _SingleSizeImageList] = {}
        images_with_size: dict[tuple[int, int], list[Image]] = {}
        if isinstance(images, ImageList):
            if images.size_count == 1:
                image_list_with_size[(images.widths[0], images.heights[0])] = images._as_single_size_image_list()
                indices_for_images_with_size[(images.widths[0], images.heights[0])] = [
                    index + current_index for index in images._as_single_size_image_list()._tensor_positions_to_indices
                ]
            else:
                for size, im_list in images._as_multi_size_image_list()._image_list_dict.items():
                    image_list_with_size[size] = im_list._as_single_size_image_list()
                    indices_for_images_with_size[size] = [
                        index + current_index
                        for index in im_list._as_single_size_image_list()._tensor_positions_to_indices
                    ]
        else:
            for image in images:
                size = (image.width, image.height)
                if size in images_with_size:
                    images_with_size[size].append(image)
                    indices_for_images_with_size[size].append(current_index)
                else:
                    images_with_size[size] = [image]
                    indices_for_images_with_size[size] = [current_index]
                current_index += 1
        image_list = self._clone()._as_multi_size_image_list()
        smallest_channel = max_channel = self.channel
        for size, ims in (images_with_size | image_list_with_size).items():
            new_indices = indices_for_images_with_size[size]
            if size in image_list._image_list_dict:
                if isinstance(ims, _SingleSizeImageList):
                    ims_tensors = [im._image_tensor for im in ims.to_images()]
                else:
                    ims_tensors = [im._image_tensor for im in ims]
                image_list._image_list_dict[size] = _SingleSizeImageList._create_image_list(
                    [im._image_tensor for im in image_list._image_list_dict[size].to_images()] + ims_tensors,
                    image_list._image_list_dict[size]._as_single_size_image_list()._tensor_positions_to_indices
                    + new_indices,
                )
            elif isinstance(ims, _SingleSizeImageList):
                if smallest_channel > ims.channel:
                    smallest_channel = ims.channel
                fixed_ims = ims
                old_indices = list(fixed_ims._indices_to_tensor_positions.items())
                fixed_ims._tensor_positions_to_indices = [
                    new_indices[i]
                    for i in sorted(
                        range(len(new_indices)),
                        key=sorted(range(len(new_indices)), key=old_indices.__getitem__).__getitem__,
                    )
                ]
                fixed_ims._indices_to_tensor_positions = fixed_ims._calc_new_indices_to_tensor_positions()
                image_list._image_list_dict[size] = fixed_ims
            else:
                image_list._image_list_dict[size] = _SingleSizeImageList._create_image_list(
                    [im._image_tensor for im in ims],
                    new_indices,
                )
                if smallest_channel > image_list._image_list_dict[size].channel:
                    smallest_channel = image_list._image_list_dict[size].channel
            for i in new_indices:
                image_list._indices_to_image_size_dict[i] = size
            max_channel = max(max_channel, image_list._image_list_dict[size].channel)
        if smallest_channel < max_channel:
            image_list = image_list.change_channel(max_channel)._as_multi_size_image_list()
        return image_list

    def remove_image_by_index(self, index: int | list[int]) -> ImageList:
        if isinstance(index, int):
            index = [index]

        invalid_indices = []
        for _i in index:
            if _i not in self._indices_to_image_size_dict:
                invalid_indices.append(_i)
        if len(invalid_indices) > 0:
            raise IndexOutOfBoundsError(invalid_indices)

        return self._remove_image_by_index_ignore_invalid(index)

    def _remove_image_by_index_ignore_invalid(self, index: int | list[int]) -> ImageList:
        from safeds.data.image.containers._empty_image_list import _EmptyImageList
        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        if isinstance(index, int):
            index = [index]
        image_list = self._clone_without_image_dict()

        for image_list_key, image_list_original in self._image_list_dict.items():
            new_single_size_image_list = (
                image_list_original._as_single_size_image_list()._remove_image_by_index_ignore_invalid(index)
            )
            if isinstance(new_single_size_image_list, _SingleSizeImageList):
                image_list._image_list_dict[image_list_key] = new_single_size_image_list
        [image_list._indices_to_image_size_dict.pop(i, None) for i in index]

        if len(image_list._image_list_dict) == 0:
            return _EmptyImageList()
        elif len(image_list._image_list_dict) == 1:
            return next(iter(image_list._image_list_dict.values()))
        else:
            return image_list

    def remove_images_with_size(self, width: int, height: int) -> ImageList:
        import torch

        _init_default_device()

        _check_remove_images_with_size_errors(width, height)
        if (width, height) not in self._image_list_dict:
            return self
        if len(self._image_list_dict) == 2:
            single_size_image_list = (
                self._image_list_dict[
                    next(iter([key for key in list(self._image_list_dict.keys()) if key != (width, height)]))
                ]
                ._clone()
                ._as_single_size_image_list()
            )
            single_size_image_list._tensor_positions_to_indices = torch.sort(
                torch.Tensor(single_size_image_list._tensor_positions_to_indices),
            )[1].tolist()
            single_size_image_list._indices_to_tensor_positions = (
                single_size_image_list._calc_new_indices_to_tensor_positions()
            )
            return single_size_image_list

        image_list = _MultiSizeImageList()
        for image_list_key, image_list_original in self._image_list_dict.items():
            if (width, height) != image_list_key:
                image_list._image_list_dict[image_list_key] = image_list_original
        for index, size in self._indices_to_image_size_dict.items():
            if size != (width, height):
                image_list._indices_to_image_size_dict[index] = size
        return image_list

    def remove_duplicate_images(self) -> ImageList:
        image_list = _MultiSizeImageList()
        indices_to_remove = []
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list_duplicates_removed = image_list_original.remove_duplicate_images()
            indices_to_remove += [
                index
                for index in image_list_original._as_single_size_image_list()._tensor_positions_to_indices
                if index not in image_list_duplicates_removed._as_single_size_image_list()._tensor_positions_to_indices
            ]
            image_list._image_list_dict[image_list_key] = image_list_duplicates_removed
        for index_rem in indices_to_remove:
            for image_list_duplicates_removed in image_list._image_list_dict.values():
                image_list_duplicates_removed._as_single_size_image_list()._tensor_positions_to_indices = [
                    index - 1 if index > index_rem else index
                    for index in image_list_duplicates_removed._as_single_size_image_list()._tensor_positions_to_indices
                ]
                image_list_duplicates_removed._as_single_size_image_list()._indices_to_tensor_positions = (
                    image_list_duplicates_removed._as_single_size_image_list()._calc_new_indices_to_tensor_positions()
                )
        next_index = 0
        for index_key, index_value in sorted(self._indices_to_image_size_dict.items()):
            if index_key not in indices_to_remove:
                image_list._indices_to_image_size_dict[next_index] = index_value
                next_index += 1
        return image_list

    def shuffle_images(self) -> ImageList:
        image_list = _MultiSizeImageList()
        new_indices = list(self._indices_to_image_size_dict.keys())
        random.shuffle(new_indices)
        current_index = 0
        for image_list_key, image_list_original in self._image_list_dict.items():
            new_image_list = image_list_original._clone()._as_single_size_image_list()
            new_image_list._tensor_positions_to_indices = new_indices[
                current_index : current_index + len(image_list_original)
            ]
            new_image_list._indices_to_tensor_positions = new_image_list._calc_new_indices_to_tensor_positions()
            image_list._image_list_dict[image_list_key] = new_image_list
            for i in new_indices[current_index : current_index + len(image_list_original)]:
                image_list._indices_to_image_size_dict[i] = image_list_key
            current_index += len(image_list_original)
        return image_list

    def resize(self, new_width: int, new_height: int) -> ImageList:
        import torch

        _init_default_device()

        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        _check_resize_errors(new_width, new_height)

        image_list_indices = []
        image_list = _SingleSizeImageList()
        image_list._tensor = torch.empty(len(self), self.channel, new_height, new_width, dtype=torch.uint8)
        current_start_index = 0
        for image_list_original in self._image_list_dict.values():
            image_list_new = image_list_original.resize(new_width, new_height)._as_single_size_image_list()
            end = current_start_index + len(image_list_original)
            image_list._tensor[current_start_index:end] = image_list_new._tensor
            image_list_indices += image_list_new._tensor_positions_to_indices
            current_start_index = end
        image_list._tensor_positions_to_indices = image_list_indices
        image_list._indices_to_tensor_positions = image_list._calc_new_indices_to_tensor_positions()
        return image_list

    def convert_to_grayscale(self) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.convert_to_grayscale()
        return image_list

    def crop(self, x: int, y: int, width: int, height: int) -> ImageList:
        import torch

        _init_default_device()

        from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList

        _check_crop_errors(x, y, width, height)

        image_list_indices = []
        image_list = _SingleSizeImageList()
        image_list._tensor = torch.empty(len(self), self.channel, width, height, dtype=torch.uint8)
        current_start_index = 0
        for image_list_original in self._image_list_dict.values():
            image_list_new = image_list_original.crop(x, y, width, height)._as_single_size_image_list()
            end = current_start_index + len(image_list_original)
            image_list._tensor[current_start_index:end] = image_list_new._tensor
            image_list_indices += image_list_new._tensor_positions_to_indices
            current_start_index = end
        image_list._tensor_positions_to_indices = image_list_indices
        image_list._indices_to_tensor_positions = image_list._calc_new_indices_to_tensor_positions()
        return image_list

    def flip_vertically(self) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.flip_vertically()
        return image_list

    def flip_horizontally(self) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.flip_horizontally()
        return image_list

    def adjust_brightness(self, factor: float) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.adjust_brightness(factor)
        return image_list

    def add_noise(self, standard_deviation: float) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.add_noise(standard_deviation)
        return image_list

    def adjust_contrast(self, factor: float) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.adjust_contrast(factor)
        return image_list

    def adjust_color_balance(self, factor: float) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.adjust_color_balance(factor)
        return image_list

    def blur(self, radius: int) -> ImageList:
        _check_blur_errors_and_warnings(radius, min(*self.widths, *self.heights), plural=True)
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.blur(radius)
        return image_list

    def sharpen(self, factor: float) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.sharpen(factor)
        return image_list

    def invert_colors(self) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.invert_colors()
        return image_list

    def rotate_right(self) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.rotate_right()
        return image_list

    def rotate_left(self) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.rotate_left()
        return image_list

    def find_edges(self) -> ImageList:
        image_list = self._clone_without_image_dict()
        for image_list_key, image_list_original in self._image_list_dict.items():
            image_list._image_list_dict[image_list_key] = image_list_original.find_edges()
        return image_list



import warnings

from safeds._validation import _check_bounds, _ClosedBound


def _check_remove_images_with_size_errors(width: int, height: int) -> None:
    _check_bounds("width", width, lower_bound=_ClosedBound(1))
    _check_bounds("height", height, lower_bound=_ClosedBound(1))


def _check_resize_errors(new_width: int, new_height: int) -> None:
    _check_bounds("new_width", new_width, lower_bound=_ClosedBound(1))
    _check_bounds("new_height", new_height, lower_bound=_ClosedBound(1))


def _check_crop_warnings(
    x: int,
    y: int,
    min_width: int,
    min_height: int,
    plural: bool,
) -> None:
    if x >= min_width or y >= min_height:
        warnings.warn(
            f"The specified bounding rectangle does not contain any content of {'at least one' if plural else 'the'} image. Therefore {'these images' if plural else 'the image'} will be blank.",
            UserWarning,
            stacklevel=2,
        )


def _check_crop_errors(
    x: int,
    y: int,
    width: int,
    height: int,
) -> None:
    _check_bounds("x", x, lower_bound=_ClosedBound(0))
    _check_bounds("y", y, lower_bound=_ClosedBound(0))
    _check_bounds("width", width, lower_bound=_ClosedBound(1))
    _check_bounds("height", height, lower_bound=_ClosedBound(1))


def _check_adjust_brightness_errors_and_warnings(factor: float, plural: bool) -> None:
    _check_bounds("factor", factor, lower_bound=_ClosedBound(0))
    if factor == 1:
        warnings.warn(
            f"Brightness adjustment factor is 1.0, this will not make changes to the {'images' if plural else 'image'}.",
            UserWarning,
            stacklevel=2,
        )


def _check_add_noise_errors(standard_deviation: float) -> None:
    _check_bounds("standard_deviation", standard_deviation, lower_bound=_ClosedBound(0))


def _check_adjust_contrast_errors_and_warnings(factor: float, plural: bool) -> None:
    _check_bounds("factor", factor, lower_bound=_ClosedBound(0))
    if factor == 1:
        warnings.warn(
            f"Contrast adjustment factor is 1.0, this will not make changes to the {'images' if plural else 'image'}.",
            UserWarning,
            stacklevel=2,
        )


def _check_adjust_color_balance_errors_and_warnings(factor: float, channel: int, plural: bool) -> None:
    _check_bounds("factor", factor, lower_bound=_ClosedBound(0))
    if factor == 1:
        warnings.warn(
            f"Color adjustment factor is 1.0, this will not make changes to the {'images' if plural else 'image'}.",
            UserWarning,
            stacklevel=2,
        )
    elif channel == 1:
        warnings.warn(
            "Color adjustment will not have an affect on grayscale images with only one channel.",
            UserWarning,
            stacklevel=2,
        )


def _check_blur_errors_and_warnings(radius: int, max_radius: int, plural: bool) -> None:
    _check_bounds("radius", radius, lower_bound=_ClosedBound(0), upper_bound=_ClosedBound(max_radius - 1))
    if radius == 0:
        warnings.warn(
            f"Blur radius is 0, this will not make changes to the {'images' if plural else 'image'}.",
            UserWarning,
            stacklevel=2,
        )


def _check_sharpen_errors_and_warnings(factor: float, plural: bool) -> None:
    _check_bounds("factor", factor, lower_bound=_ClosedBound(0))
    if factor == 1:
        warnings.warn(
            f"Sharpen factor is 1.0, this will not make changes to the {'images' if plural else 'image'}.",
            UserWarning,
            stacklevel=2,
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Self

from safeds._utils import _structural_hash

if TYPE_CHECKING:
    from safeds.data.tabular.containers import Table


class TableTransformer(ABC):
    """
    Learn a transformation for a set of columns in a `Table` and transform another `Table` with the same columns.

    Parameters
    ----------
    column_names:
        The list of columns used to fit the transformer. If `None`, all suitable columns are used.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    # The decorator is needed so the class really cannot be instantiated
    @abstractmethod
    def __init__(self, column_names: str | list[str] | None) -> None:
        if isinstance(column_names, str):
            column_names = [column_names]

        self._column_names: list[str] | None = column_names

    # The decorator ensures that the method is overridden in all subclasses
    @abstractmethod
    def __hash__(self) -> int:
        return _structural_hash(
            self.__class__.__qualname__,
            self._column_names,
            self.is_fitted,
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def is_fitted(self) -> bool:
        """Whether the transformer is fitted."""

    # ------------------------------------------------------------------------------------------------------------------
    # Learning and transformation
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def fit(self, table: Table) -> Self:
        """
        Learn a transformation for a set of columns in a table.

        **Note:** This transformer is not modified.

        Parameters
        ----------
        table:
            The table used to fit the transformer.

        Returns
        -------
        fitted_transformer:
            The fitted transformer.
        """

    @abstractmethod
    def transform(self, table: Table) -> Table:
        """
        Apply the learned transformation to a table.

        **Note:** The given table is not modified.

        Parameters
        ----------
        table:
            The table to which the learned transformation is applied.

        Returns
        -------
        transformed_table:
            The transformed table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        """

    def fit_and_transform(self, table: Table) -> tuple[Self, Table]:
        """
        Learn a transformation for a set of columns in a table and apply the learned transformation to the same table.

        **Note:** Neither this transformer nor the given table are modified.

        Parameters
        ----------
        table:
            The table used to fit the transformer. The transformer is then applied to this table.

        Returns
        -------
        fitted_transformer:
            The fitted transformer.
        transformed_table:
            The transformed table.
        """
        fitted_transformer = self.fit(table)
        transformed_table = fitted_transformer.transform(table)
        return fitted_transformer, transformed_table
